import astropy.io.fits as pyfits
import numpy as np
from soxs.utils import parse_prng, parse_value, \
    ensure_numpy_array, mylog, process_fits_string
from soxs.spatial import construct_wcs
from astropy.units import Quantity
from collections.abc import Sequence
import os
from pathlib import Path


def _parse_catalog_entry(src):
    import re
    brackets = re.findall(r"[^[]*\[([^]]*)\]", src)
    src_file = src.split("[")[0]
    ext = brackets[0].lower().split(",")
    extname = ext[0]
    if len(ext) == 2:
        extver = int(ext[1])
    else:
        extver = None
    if len(brackets) == 2 and extname == "spectrum":
        rowstr = brackets[1].lower()
        row = rowstr.split("==")[1].strip("'")
        if "row" in rowstr:
            row = int(row)
    else:
        row = None
    return src_file, extname, extver, row


class LazyReadSimputCatalog(Sequence):
    def __init__(self, src_cat):
        self.src_cat = src_cat

    def __getitem__(self, i):
        spec = self.src_cat.spectra[i]
        return self.src_cat.read_source(spec)

    def __len__(self):
        return self.src_cat.num_sources


def read_simput_catalog(simput_file):
    r"""
    Read events from a SIMPUT catalog. This will read 
    all of the sources in the catalog.

    Parameters
    ----------
    simput_file : string
        The SIMPUT file to read from.

    Returns
    -------
    1. Lists of dicts of NumPy arrays of the positions 
       and energies of the events from the sources.
    2. NumPy arrays of the parameters of the sources.
    """
    sc = SimputCatalog.from_file(simput_file)
    parameters = {"emin": sc.emin,
                  "emax": sc.emax,
                  "src_names": sc.src_names,
                  "flux": sc.fluxes}
    return LazyReadSimputCatalog(sc), parameters


class SimputCatalog:

    def __init__(self, spectra, images, src_names, ra, dec, fluxes,
                 emin, emax, filename):
        self.spectra = ensure_numpy_array(spectra)
        self.images = ensure_numpy_array(images)
        self.src_names = ensure_numpy_array(src_names)
        self.ra = ensure_numpy_array(ra)
        self.dec = ensure_numpy_array(dec)
        self.fluxes = ensure_numpy_array(fluxes)
        self.emin = ensure_numpy_array(emin)
        self.emax = ensure_numpy_array(emax)
        self.filename = filename
        self.num_sources = self.spectra.size
        # timing not yet supported
        self.timing = np.array(["NULL"]*self.num_sources)

    @classmethod
    def from_source(cls, filename, source, src_filename=None,
                    overwrite=False):
        """
        Create a new :class:`~soxs.simput.SimputCatalog`
        instance using a single :class:`~soxs.simput.SimputSource`
        instance.

        Parameters
        ----------
        filename : string
            The name of the SIMPUT catalog file to write. The
            file will be written 
        source : :class:`~soxs.simput.SimputSource`
            The SIMPUT source to create the catalog with.
        src_filename : string, optional
            If set, this will be the filename to write the source
            to. By default, the source will be written to the same
            file as the SIMPUT catalog.
        overwrite : boolean, optional
            Whether or not to overwrite an existing file with
            the same name. If src_filename=None and the source is
            to the written to the SIMPUT catalog file, then this
            argument is ignored. If src_filename is another value,
            it exists, and overwrite=False, the source will be
            appended to the file. Default: False

        """
        sc = cls([], [], [], [], [], [], [], [], filename)
        if os.path.exists(filename) and not overwrite:
            raise IOError(f"{filename} exists and overwrite=False!")
        sc._write_catalog(overwrite=overwrite)
        sc.append(source, src_filename=src_filename,
                  overwrite=overwrite)
        return sc

    @classmethod
    def from_file(cls, filename):
        """
        Generate a SIMPUT catalog object by reading it in from
        disk.

        Parameters
        ----------
        filename : string
            The name of the SIMPUT catalog file to read the
            catalog and photon lists from.
        """
        f_simput = pyfits.open(filename)
        fluxes = f_simput["src_cat"].data["flux"]
        src_names = f_simput["src_cat"].data["src_name"]
        e_min = f_simput["src_cat"].data["e_min"]
        e_max = f_simput["src_cat"].data["e_max"]
        ra = f_simput["src_cat"].data["ra"]
        dec = f_simput["src_cat"].data["dec"]
        spectra = f_simput["src_cat"].data["spectrum"]
        if "IMAGE" in f_simput["src_cat"].columns.names:
            images = f_simput["src_cat"].data["image"]
        else:
            images = ["NULL"]*len(spectra)
        f_simput.close()

        return cls(spectra, images, src_names, ra, dec, fluxes, e_min, e_max,
                   filename)

    def read_source(self, spec):
        """
        Read a source from the SIMPUT catalog with the identifier
        *spec* for the SPECTRUM field.
        """
        from .spectra import Spectrum
        from astropy.io.fits.column import _VLF
        i = np.where(self.spectra == spec)[0][0]
        src_file, extname, extver, row = _parse_catalog_entry(spec)
        # If no file is specified, assume the catalog and
        # source are in the same file
        if src_file == "":
            fn_src = self.filename
        else:
            fn_src = src_file
        ext = extname if extver is None else (extname, extver)
        data = pyfits.getdata(fn_src, ext)
        if extname == "phlist":
            ra = Quantity(data["ra"], "deg")
            dec = Quantity(data["dec"], "deg")
            energy = Quantity(data["energy"], "keV")
            src = SimputPhotonList(ra, dec, energy, self.fluxes[i],
                                   name=self.src_names[i])
        elif extname == "spectrum":
            emid = data["energy"]
            flux = data["fluxdensity"]
            if isinstance(data["energy"], _VLF):
                emid = emid[0]
                flux = flux[0]
            elif row is not None:
                if isinstance(row, str):
                    row = np.where(data["name"])[0][0]
                emid = emid[row, :]
                flux = flux[row, :]
            de = np.diff(emid)[0]
            ebins = np.append(emid - 0.5 * de, emid[-1] + 0.5 * de)
            spec = Spectrum(ebins, flux)
            if self.images[i].upper() != "NULL":
                src_file, extname, extver, _ = \
                    _parse_catalog_entry(self.images[i])
                # If no file is specified, assume the catalog and
                # source are in the same file
                if src_file == "":
                    fn_img = self.filename
                else:
                    fn_img = src_file
                ext = extname if extver is None else (extname, extver)
                imdata, imheader = pyfits.getdata(
                    fn_img, ext, header=True)
                imhdu = pyfits.ImageHDU(data=imdata,
                                        header=imheader)
            else:
                imhdu = None
            src = SimputSpectrum(spec, self.ra[i], self.dec[i],
                                 name=self.src_names[i], imhdu=imhdu)
        else:
            raise RuntimeError

        return src

    def _write_catalog(self, overwrite=False):
        """
        Write the SIMPUT catalog to disk.
        """
        src_id = np.arange(self.num_sources)
        col1 = pyfits.Column(name='SRC_ID', format='J', array=src_id)
        col2 = pyfits.Column(name='RA', format='D', array=self.ra)
        col3 = pyfits.Column(name='DEC', format='D', array=self.dec)
        col4 = pyfits.Column(name='E_MIN', format='D', array=self.emin)
        col5 = pyfits.Column(name='E_MAX', format='D', array=self.emax)
        col6 = pyfits.Column(name='FLUX', format='D', array=self.fluxes)
        col7 = pyfits.Column(name='SPECTRUM', format='80A', array=self.spectra)
        col8 = pyfits.Column(name='IMAGE', format='80A', array=self.images)
        col9 = pyfits.Column(name='TIMING', format='80A', array=self.timing)
        col10 = pyfits.Column(name='SRC_NAME', format='80A',
                              array=self.src_names)

        coldefs = pyfits.ColDefs([col1, col2, col3, col4, col5,
                                  col6, col7, col8, col9, col10])

        wrhdu = pyfits.BinTableHDU.from_columns(coldefs)
        wrhdu.name = "SRC_CAT"

        wrhdu.header["HDUCLASS"] = "HEASARC"
        wrhdu.header["HDUCLAS1"] = "SIMPUT"
        wrhdu.header["HDUCLAS2"] = "SRC_CAT"
        wrhdu.header["HDUVERS"] = "1.1.0"
        wrhdu.header["RADECSYS"] = "FK5"
        wrhdu.header["EQUINOX"] = 2000.0
        wrhdu.header["TUNIT2"] = "deg"
        wrhdu.header["TUNIT3"] = "deg"
        wrhdu.header["TUNIT4"] = "keV"
        wrhdu.header["TUNIT5"] = "keV"
        wrhdu.header["TUNIT6"] = "erg/s/cm**2"

        if os.path.exists(self.filename) and not overwrite:
            with pyfits.open(self.filename, mode='update') as f:
                f["SRC_CAT"] = wrhdu
                f.flush()
        else:
            f = [pyfits.PrimaryHDU(), wrhdu]
            pyfits.HDUList(f).writeto(self.filename, overwrite=True)

    def append(self, source, src_filename=None, overwrite=False):
        """
        Add a source to this catalog.

        Parameters
        ----------
        source : :class:`~soxs.simput.SimputSource`
            The SIMPUT source to append to this catalog.
        src_filename : string, optional
            If set, this will be the filename to write the source
            to. By default, the source will be written to the same
            file as the SIMPUT catalog.
        overwrite : boolean, optional
            Whether or not to overwrite an existing file with
            the same name. If src_filename=None and the source is
            to be written to the SIMPUT catalog file, then this
            argument is ignored. If src_filename is another value,
            it exists, and overwrite=False, the source will be
            appended to the file. Default: False
        """
        self.src_names = np.append(self.src_names, source.name)
        self.ra = np.append(self.ra, source.ra)
        self.dec = np.append(self.dec, source.dec)
        self.fluxes = np.append(self.fluxes, source.flux)
        self.emin = np.append(self.emin, source.emin)
        self.emax = np.append(self.emax, source.emax)
        self.num_sources += 1

        if src_filename is None:
            src_filename = self.filename

        if src_filename == self.filename:
            # Don't overwrite the SIMPUT catalog file!!
            overwrite = False
        elif overwrite and os.path.exists(src_filename):
            mylog.warning(f"Overwriting {src_filename}.")
            os.remove(src_filename)

        extver = _determine_extver(src_filename, source.src_type.upper())
        if src_filename != self.filename:
            src_fn = os.path.join(
                os.path.relpath(Path(src_filename).parent,
                                Path(self.filename).parent),
                os.path.basename(src_filename))
        else:
            src_fn = ""
        spec = f"{src_fn}[{source.src_type.upper()},{extver}]"
        if source.imhdu is not None:
            img_extver = _determine_extver(src_filename, "IMAGE")
            img = f"{src_fn}[IMAGE,{img_extver}]"
        elif source.src_type == "phlist":
            img = spec
            img_extver = extver
        else:
            img = "NULL"
            img_extver = None
        self.spectra = np.append(self.spectra, spec)
        self.images = np.append(self.images, img)

        self._write_catalog()
        source._write_source(src_filename, extver, img_extver=img_extver,
                             overwrite=overwrite)


def _determine_extver(fn, extname):
    extver = 1
    if os.path.exists(fn):
        with pyfits.open(fn) as f:
            for hdu in f:
                if hdu.name == extname:
                    extver = hdu.header["EXTVER"]+1
    return extver


class SimputSource:
    src_type = "null"

    def __init__(self, emin, emax, flux, ra, dec, name=None, imhdu=None):
        self.emin = emin
        self.emax = emax
        if name is None:
            name = ""
        self.name = name
        self.flux = flux
        self.ra = ra
        self.dec = dec
        self.imhdu = imhdu

    def _get_source_hdu(self):
        return None, None

    def _write_source(self, filename, extver, img_extver=None, overwrite=False):
        coldefs, header = self._get_source_hdu()

        tbhdu = pyfits.BinTableHDU.from_columns(coldefs)
        tbhdu.name = self.src_type.upper()

        if self.src_type == "phlist":
            hduclas1 = "PHOTONS"
        else:
            hduclas1 = self.src_type.upper()
        tbhdu.header["HDUCLASS"] = "HEASARC/SIMPUT"
        tbhdu.header["HDUCLAS1"] = hduclas1
        tbhdu.header["HDUVERS"] = "1.1.0"
        tbhdu.header.update(header)

        tbhdu.header["EXTVER"] = extver
        if self.imhdu is not None:
            self.imhdu.header["EXTVER"] = img_extver

        if os.path.exists(filename) and not overwrite:
            mylog.info(f"Appending this source to {filename}.")
            with pyfits.open(filename, mode='append') as f:
                f.append(tbhdu)
                if self.imhdu is not None:
                    f.append(self.imhdu)
                f.flush()
        else:
            if os.path.exists(filename):
                mylog.warning(f"Overwriting {filename} with this source.")
            else:
                mylog.info(f"Writing source to {filename}.")
            f = [pyfits.PrimaryHDU(), tbhdu]
            if self.imhdu is not None:
                f.append(self.imhdu)
            pyfits.HDUList(f).writeto(filename, overwrite=overwrite)

        if self.imhdu is not None:
            self.imhdu.header["EXTVER"] = 1


class SimputSpectrum(SimputSource):
    src_type = "spectrum"

    def __init__(self, spec, ra, dec, name=None, imhdu=None):
        super(SimputSpectrum, self).__init__(spec.ebins.value.min(),
                                             spec.ebins.value.max(),
                                             spec.total_flux.value, ra,
                                             dec, name=name, imhdu=imhdu)
        self.spec = spec

    def _get_source_hdu(self):
        col1 = pyfits.Column(name='ENERGY', format='E',
                             array=self.spec.emid.value)
        col2 = pyfits.Column(name='FLUXDENSITY', format='D',
                             array=self.spec.flux.value)
        cols = [col1, col2]

        coldefs = pyfits.ColDefs(cols)

        header = {"REFRA": self.ra,
                  "REFDEC": self.dec,
                  "TUNIT1": "keV",
                  "TUNIT2": "photon/(cm**2*s*keV)"}

        return coldefs, header

    @classmethod
    def from_spectrum(cls, name, spectral_model, ra, dec, imhdu=None):
        """
        Generates a SIMPUT spectrum model for a point source
        from a spectral model and a coordinate on the sky. An image
        can be optionally included for an extended source.

        Parameters
        ----------
        name : string
            The name of the SIMPUT spectrum.
        spectral_model : :class:`~soxs.spectra.Spectrum`
            The spectral model to use to generate the event energies.
        ra : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The RA of the source in degrees.
        dec : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The Dec of the source in degrees.
        imhdu : string or :class:`~astropy.io.fits.ImageHDU` instance
            An image to be used to simulate an extended source. Can be an
            ImageHDU instance or the name of a to read one from. If the
            name contains an HDU extension, e.g. "cluster.fits[1]" or
            "cluster.fits['perseus']", that extension will be loaded.
        """
        if isinstance(imhdu, str):
            imhdu = process_fits_string(imhdu)
        return cls(spectral_model, ra, dec, name=name, imhdu=imhdu)

    @classmethod
    def from_models(cls, name, spectral_model, spatial_model,
                    width, nx):
        """
        Generate a SIMPUT spectrum from a spectral and a spatial
        model, and parameters for an image.

        Parameters
        ----------
        name : string
            The name of the SIMPUT spectrum.
        spectral_model : :class:`~soxs.spectra.Spectrum`
            The spectral model to use to generate the event energies.
        spatial_model : :class:`~soxs.spatial.SpatialModel`
            The spatial model to use to generate the event coordinates.
        width : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The width of the image in arcminutes.
        nx : integer
            The resolution of the image, e.g. the number of pixels
            on a side.
        """
        imhdu = spatial_model.generate_image(width, nx)
        return cls(spectral_model, spatial_model.ra0,
                   spatial_model.dec0, name=name, imhdu=imhdu)


class SimputPhotonList(SimputSource):
    src_type = "phlist"

    def __init__(self, ra, dec, energy, flux, name=None):
        emin = np.asarray(energy).min()
        emax = np.asarray(energy).max()
        super(SimputPhotonList, self).__init__(
            emin, emax, flux, 0.0, 0.0, name=name)
        self.events = {"ra": ra, "dec": dec, "energy": energy}
        self.num_events = energy.size

    def __getitem__(self, item):
        return self.events[item]

    def __contains__(self, item):
        return item in self.events

    def __iter__(self):
        for key in self.events:
            yield key

    @classmethod
    def from_models(cls, name, spectral_model, spatial_model,
                    t_exp, area, prng=None):
        """
        Generate a SIMPUT photon list from a spectral and a spatial
        model. 

        Parameters
        ----------
        name : string
            The name of the SIMPUT photon list.
        spectral_model : :class:`~soxs.spectra.Spectrum`
            The spectral model to use to generate the event energies.
        spatial_model : :class:`~soxs.spatial.SpatialModel`
            The spatial model to use to generate the event coordinates.
        t_exp : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The exposure time in seconds.
        area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The effective area in cm**2. If one is creating 
            events for a SIMPUT file, a constant should be 
            used and it must be large enough so that a 
            sufficiently large sample is drawn for the ARF.
        prng : :class:`~numpy.random.RandomState` object, integer, or None
            A pseudo-random number generator. Typically will only 
            be specified if you have a reason to generate the same 
            set of random numbers, such as for a test. Default is None, 
            which sets the seed based on the system time. 
        """
        prng = parse_prng(prng)
        t_exp = parse_value(t_exp, "s")
        area = parse_value(area, "cm**2")
        e = spectral_model.generate_energies(t_exp, area, prng=prng)
        ra, dec = spatial_model.generate_coords(e.size, prng=prng)
        return cls(ra, dec, e, e.flux.value, name=name)

    def _get_source_hdu(self):
        col1 = pyfits.Column(name='ENERGY', format='E',
                             array=np.asarray(self["energy"]))
        col2 = pyfits.Column(name='RA', format='D',
                             array=np.asarray(self["ra"]))
        col3 = pyfits.Column(name='DEC', format='D',
                             array=np.asarray(self["dec"]))
        cols = [col1, col2, col3]

        coldefs = pyfits.ColDefs(cols)

        header = {"REFRA": 0.0,
                  "REFDEC": 0.0,
                  "TUNIT1": "keV",
                  "TUNIT2": "deg",
                  "TUNIT3": "deg"}

        return coldefs, header

    def plot(self, center, width, s=None, c=None, marker=None, stride=1,
             emin=None, emax=None, label=None, fontsize=18, fig=None, 
             ax=None, **kwargs):
        """
        Plot event coordinates from this photon list in a scatter plot, 
        optionally restricting the photon energies which are plotted
        and using only a subset of the photons. 

        Parameters
        ----------
        center : array-like
            The RA, Dec of the center of the plot in degrees.
        width : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The width of the plot in arcminutes.
        s : integer, optional
            Size of the scatter marker in points^2.
        c : string, optional
            The color of the points.
        marker : string, optional
            The marker to use for the points in the scatter plot. Default: 'o'
        stride : integer, optional
            Plot every *stride* events. Default: 1
        emin : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The minimum energy of the photons to plot. Default is
            the minimum energy in the list.
        emax : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The maximum energy of the photons to plot. Default is
            the maximum energy in the list.
        label : string, optional
            The label of the spectrum. Default: None
        fontsize : int
            Font size for labels and axes. Default: 18
        fig : :class:`~matplotlib.figure.Figure`, optional
            A Figure instance to plot in. Default: None, one will be
            created if not provided.
        ax : :class:`~matplotlib.axes.Axes`, optional
            An Axes instance to plot in. Default: None, one will be
            created if not provided.
        """
        import matplotlib.pyplot as plt
        from astropy.visualization.wcsaxes import WCSAxes
        if fig is None:
            fig = plt.figure(figsize=(10, 10))
        if ax is None:
            wcs = construct_wcs(center[0], center[1])
            ax = WCSAxes(fig, [0.15, 0.1, 0.8, 0.8], wcs=wcs)
            fig.add_axes(ax)
        else:
            wcs = ax.wcs
        if emin is None:
            emin = self.emin
        else:
            emin = parse_value(emin, "keV")
        if emax is None:
            emax = self.emax
        else:
            emax = parse_value(emax, "keV")
        idxs = np.logical_and(self["energy"].value >= emin, 
                              self["energy"].value <= emax)
        ra = self["ra"][idxs][::stride].value
        dec = self["dec"][idxs][::stride].value
        x, y = wcs.wcs_world2pix(ra, dec, 1)
        ax.scatter(x, y, s=s, c=c, marker=marker, label=label, **kwargs)
        x0, y0 = wcs.wcs_world2pix(center[0], center[1], 1)
        width = parse_value(width, "arcmin")*60.0
        ax.set_xlim(x0-0.5*width, x0+0.5*width)
        ax.set_ylim(y0-0.5*width, y0+0.5*width)
        ax.set_xlabel("RA")
        ax.set_ylabel("Dec")
        ax.tick_params(axis='both', labelsize=fontsize)
        return fig, ax


def write_photon_list(simput_prefix, phlist_prefix, flux, ra, dec, 
                      energy, overwrite=False):
    """
    This function is designed to preserve backwards-compatibility
    with pyXSIM 2.x. It will be removed in a future release.
    """
    simput_file = f"{simput_prefix}_simput.fits"
    phlist_file = f"{phlist_prefix}_phlist.fits"
    phlist = SimputPhotonList(ra, dec, energy, flux)
    SimputCatalog.from_source(simput_file, phlist, 
                              src_filename=phlist_file,
                              overwrite=overwrite)


from soxs.spectra import Spectrum, ConvolvedSpectrum
from soxs.response import AuxiliaryResponseFile
from numpy.testing import assert_allclose, assert_array_equal
import os
import tempfile
import shutil


def test_arithmetic():
    spec1 = Spectrum.from_powerlaw(1.0, 0.05, 1.0e-4, 0.1, 10.0, 10000)
    spec2 = Spectrum.from_powerlaw(2.0, 0.01, 1.0e-3, 0.1, 10.0, 10000)
    spec3 = spec1+spec2
    spec3b = spec1-spec2
    flux3 = spec1.flux+spec2.flux
    flux3b = spec1.flux-spec2.flux

    assert_allclose(spec3.flux, flux3)
    assert_allclose(spec3b.flux, flux3b)

    spec4 = spec3*3.0
    spec5 = 3.0*spec3

    flux4 = spec3.flux*3.0

    assert_allclose(spec4.flux, spec5.flux)
    assert_allclose(spec4.flux, flux4)

    spec6 = spec3/2.5
    flux6 = spec3.flux/2.5

    assert_allclose(spec6.flux, flux6)

    spec7 = Spectrum.from_constant(1.0e-4, 0.1, 10.0, 10000)
    spec8 = spec1+spec7

    assert_allclose(spec8.flux.value, spec1.flux.value+1.0e-4)

    spec8 += spec1

    assert_allclose(spec8.flux.value, (2.0*spec1+spec7).flux.value)


def test_read_write():

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    spec1 = Spectrum.from_powerlaw(1.0, 0.05, 1.0e-4, 0.1, 10.0, 10000)
    spec1.write_file("test_spec.dat", overwrite=True)
    spec2 = Spectrum.from_file("test_spec.dat")

    assert_allclose(spec1.flux, spec2.flux)
    assert_allclose(spec1.emid, spec2.emid)
    assert_allclose(spec1.ebins, spec2.ebins)
    assert_allclose(spec1.cumspec, spec2.cumspec)

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_rescale_flux():
    spec = Spectrum.from_powerlaw(2.0, 0.01, 1.0, 0.1, 10.0, 10000)

    spec.rescale_flux(1.0e-4, emin=0.5, emax=7.0, flux_type="photons")
    f = spec.get_flux_in_band(0.5, 7.0)[0]
    assert_allclose(1.0e-4, f.value)

    spec.rescale_flux(1.0e-12, emin=0.4, emax=1.0, flux_type="energy")
    f = spec.get_flux_in_band(0.4, 1.0)[1]
    assert_allclose(1.0e-12, f.value)


def test_convolved_spectra():
    arf = AuxiliaryResponseFile("xrs_hdxi_3x10.arf")
    spec1 = Spectrum.from_powerlaw(2.0, 0.01, 1.0, 0.1, 10.0, 1000)
    cspec1 = ConvolvedSpectrum.convolve(spec1, arf)
    cspec2 = spec1*arf
    spec2 = cspec1.deconvolve()
    assert_array_equal(cspec1.ebins.value, cspec2.ebins.value)
    assert_array_equal(spec1.ebins.value, spec2.ebins.value)
    assert_array_equal(cspec1.flux.value, cspec2.flux.value)
    assert_allclose(spec1.flux.value, spec2.flux.value)


import numpy as np
import os
import astropy.io.fits as pyfits
from soxs.utils import mylog, get_rot_mat

key_map = {"telescope": "TELESCOP",
           "mission": "MISSION",
           "instrument": "INSTRUME",
           "channel_type": "CHANTYPE",
           "nchan": "PHA_BINS"}


def add_background_from_file(events, event_params, bkg_file):
    from soxs.instrument import perform_dither
    f = pyfits.open(bkg_file)

    hdu = f["EVENTS"]

    dither_params = {}
    if "DITHXAMP" in hdu.header:
        dither_params["x_amp"] = hdu.header["DITHXAMP"]
        dither_params["y_amp"] = hdu.header["DITHYAMP"]
        dither_params["x_period"] = hdu.header["DITHXPER"]
        dither_params["y_period"] = hdu.header["DITHYPER"]
        dither_params["plate_scale"] = hdu.header["TCDLT3"]*3600.0
        dither_params["dither_on"] = True
    else:
        dither_params["dither_on"] = False

    sexp = event_params["exposure_time"]
    bexp = hdu.header["EXPOSURE"]

    if event_params["exposure_time"] > hdu.header["EXPOSURE"]:
        raise RuntimeError(f"The background file does not have sufficient "
                           f"exposure! Source exposure time {sexp}, background "
                           f" exposure time {bexp}.")

    for k1, k2 in key_map.items():
        if event_params[k1] != hdu.header[k2]:
            raise RuntimeError(f"'{k1}' keyword does not match! "
                               f"{event_params[k1]} vs. {hdu.header[k2]}")
    rmf1 = os.path.split(event_params["rmf"])[-1]
    rmf2 = hdu.header["RESPFILE"]
    arf1 = os.path.split(event_params["arf"])[-1]
    arf2 = hdu.header["ANCRFILE"]
    if rmf1 != rmf2:
        raise RuntimeError(f"RMFs do not match! {rmf1} vs. {rmf2}")
    if arf1 != arf2:
        raise RuntimeError(f"ARFs do not match! {arf1} vs. {arf2}")

    idxs = hdu.data["TIME"] < sexp

    mylog.info(f"Adding {idxs.sum()} background events from {bkg_file}.")

    if event_params["roll_angle"] == hdu.header["ROLL_PNT"]:
        xpix = hdu.data["X"][idxs]
        ypix = hdu.data["Y"][idxs]
    else:
        rot_mat = get_rot_mat(event_params["roll_angle"])
        if dither_params["dither_on"]:
            t = hdu.data["TIME"][idxs]
            x_off, y_off = perform_dither(t, dither_params)
        else:
            x_off = 0.0
            y_off = 0.0
        det = np.array([hdu.data["DETX"][idxs] + x_off -
                        event_params["aimpt_coords"][0] -
                        event_params["aimpt_shift"][0],
                        hdu.data["DETY"][idxs] + y_off -
                        event_params["aimpt_coords"][1] -
                        event_params["aimpt_shift"][1]])
        xpix, ypix = np.dot(rot_mat.T, det)

        xpix += hdu.header["TCRPX2"]
        ypix += hdu.header["TCRPX3"]

    all_events = {}
    for key in ["detx", "dety", "time", "ccd_id", event_params["channel_type"]]:
        all_events[key] = np.concatenate([events[key], 
                                          hdu.data[key.upper()][idxs]])
    all_events["xpix"] = np.concatenate([events["xpix"], xpix])
    all_events["ypix"] = np.concatenate([events["ypix"], ypix])
    all_events["energy"] = np.concatenate([events["energy"],
                                           hdu.data["ENERGY"][idxs]*1.0e-3])

    f.close()

    return all_events


def make_diffuse_background(bkg_events, event_params, rmf, prng=None):
    from soxs.instrument import perform_dither

    n_e = bkg_events["energy"].size

    bkg_events['time'] = prng.uniform(size=n_e, low=0.0,
                                      high=event_params["exposure_time"])

    x_offset, y_offset = perform_dither(bkg_events["time"],
                                        event_params["dither_params"])

    rot_mat = get_rot_mat(event_params["roll_angle"])

    det = np.array([bkg_events["detx"] + x_offset -
                    event_params["aimpt_coords"][0] -
                    event_params["aimpt_shift"][0],
                    bkg_events["dety"] + y_offset -
                    event_params["aimpt_coords"][1] -
                    event_params["aimpt_shift"][1]])
    pix = np.dot(rot_mat.T, det)

    bkg_events["xpix"] = pix[0, :] + event_params['pix_center'][0]
    bkg_events["ypix"] = pix[1, :] + event_params['pix_center'][1]

    return bkg_events


import pytest
import os


def pytest_addoption(parser):
    parser.addoption("--answer_dir",
                     help="Directory where answers are stored.")
    parser.addoption("--answer_store", action="store_true",
                     help="Generate new answers, but don't test.")


@pytest.fixture()
def answer_store(request):
    return request.config.getoption('--answer_store')


@pytest.fixture()
def answer_dir(request):
    ad = os.path.abspath(request.config.getoption('--answer_dir'))
    if not os.path.exists(ad):
        os.makedirs(ad)
    return ad



import os
import numpy as np
import h5py

from tqdm.auto import tqdm

from astropy.cosmology import FlatLambdaCDM
from astropy.table import Table

from soxs.spatial import BetaModel, construct_wcs
from soxs.spectra import ApecGenerator
from soxs.utils import soxs_files_path, mylog, parse_prng, \
    parse_value
from soxs.simput import SimputCatalog, SimputPhotonList

# Cosmological parameters for the catalog
# SHOULD NOT BE ALTERED
omega_m = 0.279
omega_l = 0.721
omega_k = 1.0 - omega_m - omega_l
h0 = 0.7

# Parameters for the halo fluxes
# SHOULD NOT BE ALTERED
emin = 0.5
emax = 2.0
abund = 0.3
conc = 10.0

lum_table_file = os.path.join(soxs_files_path, "lum_table.h5")
halos_cat_file = os.path.join(soxs_files_path, "halo_catalog.h5")


def lum(M_mean, z_mean): 
    # Alexey's cosmology II paper, eq. 22
    E_z = np.sqrt(omega_m*(1.+z_mean)**3+omega_k*(1.+z_mean)**2+omega_l)
    ln_Lx = 47.392+1.61*np.log(M_mean)+1.850*np.log(E_z)-0.39*np.log(h0/0.72)
    return np.exp(ln_Lx)


def Tx(M_mean, z_mean): 
    # Alexey's cosmology II paper, eq. 6, Table 3 for coefficients.
    E_z = np.sqrt(omega_m*(1.+z_mean)**3+omega_k*(1.+z_mean)**2+omega_l)
    return 5.0 * (M_mean*E_z*h0/3.02e14)**(1.0/1.53)


def flux2lum(kT, z):
    lum_table = h5py.File(lum_table_file, "r")
    kT_idxs = np.round((kT-0.1)/0.1).astype('int')
    z_idxs = np.round(z/0.05).astype('int')
    flux2lum = lum_table["Lx"][()][kT_idxs, z_idxs]
    lum_table.close()
    return flux2lum


def make_cosmological_sources(exp_time, fov, sky_center, cat_center=None,
                              absorb_model="wabs", nH=0.05, area=40000.0,
                              output_sources=None, write_regions=None,
                              prng=None):
    r"""
    Make an X-ray source made up of contributions from
    galaxy clusters, galaxy groups, and galaxies. 

    Parameters
    ----------
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time of the observation in seconds.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    cat_center : array-like
        The center of the field in the coordinates of the
        halo catalog, which range from -5.0 to 5.0 in 
        degrees in both directions. If None is given, a 
        center will be randomly chosen.
    absorb_model : string, optional
        The absorption model to use, "wabs" or "tbabs". Default: "wabs"
    nH : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The hydrogen column in units of 10**22 atoms/cm**2. 
        Default: 0.05
    area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The effective area in cm**2. It must be large enough 
        so that a sufficiently large sample is drawn for the 
        ARF. Default: 40000.
    output_sources : string, optional
        If set to a filename, output the properties of the sources
        within the field of view to a file. Default: None
    write_regions : string, optional
        If set to a filename, output circle ds9 regions corresponding to the
        positions of the halos with radii corresponding to their R500 
        projected on the sky.  Default: None
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time.
    """
    exp_time = parse_value(exp_time, "s")
    fov = parse_value(fov, "arcmin")
    if nH is not None:
        nH = parse_value(nH, "1.0e22*cm**-2")
    area = parse_value(area, "cm**2")
    prng = parse_prng(prng)
    cosmo = FlatLambdaCDM(H0=100.0*h0, Om0=omega_m)
    agen = ApecGenerator(0.1, 10.0, 10000, broadening=False)

    mylog.info("Creating photons from cosmological sources.")

    mylog.info("Loading halo data from catalog: %s" % halos_cat_file)
    halo_data = h5py.File(halos_cat_file, "r")

    scale = cosmo.kpc_comoving_per_arcmin(halo_data["redshift"][()]).to("Mpc/arcmin")

    # 600. arcmin = 10 degrees (total FOV of catalog = 100 deg^2)
    fov_cat = 10.0*60.0
    w = construct_wcs(*sky_center)

    cat_min = -0.5*fov_cat
    cat_max = 0.5*fov_cat

    if cat_center is None:
        xc, yc = prng.uniform(low=cat_min+0.5*fov, high=cat_max-0.5*fov, size=2)
    else:
        xc, yc = cat_center
        xc *= 60.0
        yc *= 60.0
        xc, yc = np.clip([xc, yc], cat_min+0.5*fov, cat_max-0.5*fov)

    mylog.info("Coordinates of the FOV within the catalog are (%g, %g) deg." %
               (xc/60.0, yc/60.0))

    xlo = (xc-1.1*0.5*fov)
    xhi = (xc+1.1*0.5*fov)
    ylo = (yc-1.1*0.5*fov)
    yhi = (yc+1.1*0.5*fov)

    mylog.info("Selecting halos in the FOV.")

    halo_x = halo_data["x"][()].astype("float64")/(h0*scale.value)
    halo_y = halo_data["y"][()].astype("float64")/(h0*scale.value)

    fov_idxs = (halo_x >= xlo) & (halo_x <= xhi)
    fov_idxs = (halo_y >= ylo) & (halo_y <= yhi) & fov_idxs

    n_halos = fov_idxs.sum()

    mylog.info("Number of halos in the field of view: %d" % n_halos)

    # Now select the specific halos which are in the FOV
    z = halo_data["redshift"][fov_idxs].astype("float64")
    m = halo_data["M500c"][fov_idxs].astype("float64")/h0
    # We need to compute proper scales here
    s = scale[fov_idxs].to("Mpc/arcsec").value/(1.0+z)
    ra0, dec0 = w.wcs_pix2world((halo_x[fov_idxs]-xc)*60.0,
                                (halo_y[fov_idxs]-yc)*60.0, 1)

    # Close the halo catalog file
    halo_data.close()

    # Some cosmological stuff
    rho_crit = cosmo.critical_density(z).to("Msun/Mpc**3").value

    # halo temperature and k-corrected flux
    kT = Tx(m, z)
    flux_kcorr = 1.0e-14*lum(m, z)/flux2lum(kT, z)

    # halo scale radius
    r500 = (3.0*m/(4.0*np.pi*500*rho_crit))**(1.0/3.0)
    r500_kpc = r500 * 1000.0
    rc_kpc = r500/conc * 1000.0
    rc = r500/conc/s

    # Halo slope parameter
    beta = prng.normal(loc=0.666, scale=0.05, size=n_halos)
    beta[beta < 0.5] = 0.5

    # Halo ellipticity
    ellip = prng.normal(loc=0.85, scale=0.15, size=n_halos)
    ellip[ellip < 0.0] = 1.0e-3

    # Halo orientation
    theta = 360.0*prng.uniform(size=n_halos)

    # If requested, output the source properties to a file
    if output_sources is not None:
        t = Table([ra0, dec0, rc_kpc, beta, ellip, theta, m,
                   r500_kpc, kT, z, flux_kcorr],
                  names=('RA', 'Dec', 'r_c', 'beta', 'ellipticity',
                         'theta', 'M500c', 'r500', 'kT', 'redshift',
                         'flux_0.5_2.0_keV'))
        t["RA"].unit = "deg"
        t["Dec"].unit = "deg"
        t["flux_0.5_2.0_keV"].unit = "erg/(cm**2*s)"
        t["r_c"].unit = "kpc"
        t["theta"].unit = "deg"
        t["M500c"].unit = "solMass"
        t["r500"].unit = "kpc"
        t["kT"].unit = "kT"
        t.write(output_sources, format='ascii.ecsv', overwrite=True)

    if write_regions is not None:
        from regions import CircleSkyRegion, write_ds9
        from astropy.coordinates import Angle, SkyCoord
        regs = []
        for halo in range(n_halos):
            c = SkyCoord(ra0[halo], dec0[halo], unit=("deg", "deg"), frame='fk5')
            scale = cosmo.kpc_proper_per_arcmin(z[halo]).to("kpc/deg")
            r500c = r500_kpc / scale.value
            r = Angle(r500c, 'deg')
            reg = CircleSkyRegion(c, r)
            regs.append(reg)
        write_ds9(regs, write_regions)

    tot_flux = 0.0
    ee = []
    ra = []
    dec = []

    pbar = tqdm(leave=True, total=n_halos, 
                desc="Generating photons from halos ")
    for halo in range(n_halos):
        spec = agen.get_spectrum(kT[halo], abund, z[halo], 1.0)
        spec.rescale_flux(flux_kcorr[halo], emin=emin, emax=emax, 
                          flux_type="energy")
        if nH is not None:
            spec.apply_foreground_absorption(nH, model=absorb_model)
        e = spec.generate_energies(exp_time, area, prng=prng, quiet=True)
        beta_model = BetaModel(ra0[halo], dec0[halo], rc[halo], beta[halo],
                               ellipticity=ellip[halo], theta=theta[halo])
        xsky, ysky = beta_model.generate_coords(e.size, prng=prng)
        tot_flux += e.flux
        ee.append(e.value)
        ra.append(xsky.value)
        dec.append(ysky.value)
        pbar.update()
    pbar.close()

    ra = np.concatenate(ra)
    dec = np.concatenate(dec)
    ee = np.concatenate(ee)

    mylog.info("Created %d photons from cosmological sources." % ee.size)

    output_events = {"ra": ra, "dec": dec, "energy": ee, 
                     "flux": tot_flux.value}

    return output_events


def make_cosmological_sources_file(filename, name, exp_time, fov,
                                   sky_center, cat_center=None,
                                   absorb_model="wabs", nH=0.05, area=40000.0,
                                   overwrite=False, output_sources=None, 
                                   write_regions=None, src_filename=None,
                                   prng=None, append=False):
    r"""
    Make a SIMPUT catalog made up of contributions from
    galaxy clusters, galaxy groups, and galaxies.

    Parameters
    ----------
    filename : string
        The filename for the SIMPUT catalog.
    name : string
        The name of the SIMPUT photon list.
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time of the observation in seconds.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    cat_center : array-like
        The center of the field in the coordinates of the
        halo catalog, which range from -5.0 to 5.0 degrees
        along both axes. If None is given, a center will be
        randomly chosen.
    absorb_model : string, optional
        The absorption model to use, "wabs" or "tbabs". Default: "wabs"
    nH : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The hydrogen column in units of 10**22 atoms/cm**2. 
        Default: 0.05
    area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The effective area in cm**2. It must be large enough 
        so that a sufficiently large sample is drawn for the 
        ARF. Default: 40000.
    overwrite : boolean, optional
        Set to True to overwrite previous files. Default: False
    output_sources : string, optional
        If set to a filename, output the properties of the sources
        within the field of view to an ASCII file. Default: None
    write_regions : string, optional
        If set to a filename, output circle ds9 regions corresponding to the
        positions of the halos with radii corresponding to their R500 
        projected on the sky. Default: None
    src_filename : string, optional
        If set, this will be the filename to write the source
        to. By default, the source will be written to the same
        file as the SIMPUT catalog
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    append : boolean, optional
        If True, the photon list source will be appended to an existing
        SIMPUT catalog. Default: False
    """
    events = make_cosmological_sources(exp_time, fov, sky_center,
                                       cat_center=cat_center,
                                       absorb_model=absorb_model, nH=nH,
                                       area=area, output_sources=output_sources,
                                       write_regions=write_regions,prng=prng)
    phlist = SimputPhotonList(events["ra"], events["dec"], events["energy"],
                              events["flux"], name=name)
    if append:
        cat = SimputCatalog.from_file(filename)
        cat.append(phlist, src_filename=src_filename, overwrite=overwrite)
    else:
        cat = SimputCatalog.from_source(filename, phlist,
                                        src_filename=src_filename,
                                        overwrite=overwrite)
    return cat


import os
from numpy.testing import assert_allclose, assert_equal
from astropy.io import fits
import shutil
import numpy as np


def spectrum_answer_testing(spec, filename, answer_store, answer_dir):
    testfile = os.path.join(answer_dir, filename)
    if answer_store:
        spec.write_h5_file(testfile, overwrite=True)
    else:
        answer_spec = type(spec).from_file(testfile)
        assert_allclose(answer_spec.emid.value,
                             spec.emid.value)
        assert_allclose(answer_spec.flux.value,
                             spec.flux.value)
        assert answer_spec.flux.unit == spec.flux.unit


def file_answer_testing(hdu, filename, answer_store, answer_dir):
    oldf = os.path.join(answer_dir, filename)
    if answer_store:
        shutil.copy(filename, answer_dir)
    else:
        f_old = fits.open(oldf)
        f_new = fits.open(filename)
        if hdu in ["IMAGE", "EXPMAP"]:
            for k in f_old[hdu].header:
                assert_equal(f_old[hdu].header[k], f_new[hdu].header[k])
            dtype = f_old[hdu].data.dtype
            if np.issubdtype(dtype, np.float32):
                rtol = 1.0e-6
            else:
                rtol = 1.0e-8
            assert_allclose(f_old[hdu].data, f_new[hdu].data, rtol=rtol)
        else:
            old_cols = f_old[hdu].data.names
            new_cols = f_new[hdu].data.names
            assert old_cols == new_cols
            for name in old_cols:
                dtype = f_old[hdu].data[name].dtype
                if np.issubdtype(dtype, np.integer):
                    assert_equal(f_old[hdu].data[name], f_new[hdu].data[name])
                else:
                    if np.issubdtype(dtype, np.float32):
                        rtol = 1.0e-6
                    else:
                        rtol = 1.0e-8
                    assert_allclose(f_old[hdu].data[name],
                                    f_new[hdu].data[name],
                                    rtol=rtol)
        f_old.close()
        f_new.close()



import os
import shutil
import tempfile
import numpy as np

from astropy.io import fits

from soxs.spectra import Spectrum
from soxs.instrument import instrument_simulator
from soxs.simput import SimputSpectrum, SimputCatalog
from soxs.events import filter_events

def test_filter():

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    alpha_sim = 1.1
    nH_sim = 0.02
    norm_sim = 1.0e-4
    redshift = 0.01

    exp_time = (100.0, "ks")
    area = 1000.0
    inst_name = "chandra_acisi_cy0"

    spec = Spectrum.from_powerlaw(alpha_sim, redshift, norm_sim, 0.1, 10.0,
                                  2000)
    spec.apply_foreground_absorption(nH_sim, model="tbabs")

    pt_src = SimputSpectrum.from_spectrum("plaw_model", spec, 30.0, 45.0)
    cat = SimputCatalog.from_source("plaw_model_simput.fits", pt_src,
                                    overwrite=True)

    instrument_simulator("plaw_model_simput.fits", "evt.fits",
                         exp_time, inst_name, [30.0, 45.0], instr_bkgnd=False,
                         ptsrc_bkgnd=False, foreground=True, prng=24)

    filter_events("evt.fits", "evt_filter_en.fits", emin=0.5, emax=2.0, overwrite=True)
    with fits.open("evt_filter_en.fits") as f:
        e = f["EVENTS"].data["ENERGY"]
        assert np.logical_and(e > 500.0, e < 2000.0).all()

    reg = "# Region file format: DS9\nimage\ncircle(2430,2454,200)\n"
    filter_events("evt.fits", "evt_filter_reg.fits", region=reg, overwrite=True)
    with fits.open("evt_filter_reg.fits") as f:
        x = f["EVENTS"].data["X"]
        y = f["EVENTS"].data["Y"]
        r = np.sqrt((x-2430)**2+(y-2454)**2)
        assert np.all(r < 201.0)

    os.chdir(curdir)
    shutil.rmtree(tmpdir)



import numpy as np
import astropy.wcs as pywcs
import os
from collections import defaultdict
from regions import PixCoord
import warnings

from soxs.events import write_event_file
from soxs.instrument_registry import instrument_registry
from soxs.psf import psf_model_registry
from soxs.response import AuxiliaryResponseFile, RedistributionMatrixFile
from soxs.simput import read_simput_catalog, SimputPhotonList
from soxs.utils import mylog, parse_prng, parse_value, \
    get_rot_mat, create_region, get_data_file, ensure_numpy_array


def perform_dither(t, dither_dict):
    if dither_dict["dither_on"]:
        a = 2.0*np.pi/dither_dict["x_period"]
        b = 2.0*np.pi/dither_dict["y_period"]
        A = dither_dict["x_amp"]/dither_dict["plate_scale"]
        B = dither_dict["y_amp"]/dither_dict["plate_scale"]
        x_offset = A*np.sin(a*t)
        y_offset = B*np.sin(b*t)
    else:
        x_offset = np.zeros(t.size)
        y_offset = np.zeros(t.size)
    return x_offset, y_offset


def generate_events(source, exp_time, instrument, sky_center, 
                    no_dither=False, dither_params=None, 
                    roll_angle=0.0, subpixel_res=False, 
                    aimpt_shift=None, prng=None):
    """
    Take unconvolved events and convolve them with instrumental responses. This 
    function does the following:

    1. Determines which events are observed using the ARF
    2. Pixelizes the events, applying PSF effects and dithering
    3. Determines energy channels using the RMF

    This function is not meant to be called by the end-user but is used by
    the :func:`~soxs.instrument.instrument_simulator` function.

    Parameters
    ----------
    input_events : string, dict, or None
        The unconvolved events to be used as input. Can be one of the
        following:
        1. The name of a SIMPUT catalog file.
        2. A Python dictionary containing the following items:
        "ra": A NumPy array of right ascension values in degrees.
        "dec": A NumPy array of declination values in degrees.
        "energy": A NumPy array of energy values in keV.
        "flux": The flux of the entire source, in units of erg/cm**2/s.
    out_file : string
        The name of the event file to be written.
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time to use, in seconds. 
    instrument : string
        The name of the instrument to use, which picks an instrument
        specification from the instrument registry. 
    sky_center : array, tuple, or list
        The center RA, Dec coordinates of the observation, in degrees.
    no_dither : boolean, optional
        If True, turn off dithering entirely. Default: False
    dither_params : array-like of floats, optional
        The parameters to use to control the size and period of the dither
        pattern. The first two numbers are the dither amplitude in x and y
        detector coordinates in arcseconds, and the second two numbers are
        the dither period in x and y detector coordinates in seconds. 
        Default: [8.0, 8.0, 1000.0, 707.0].
    roll_angle : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The roll angle of the observation in degrees. Default: 0.0
    subpixel_res : boolean, optional
        If True, event positions are not randomized within the pixels 
        within which they are detected. Default: False
    aimpt_shift : array-like, optional
        A two-float array-like object which shifts the aimpoint on the 
        detector from the nominal position. Units are in arcseconds.
        Default: None, which results in no shift from the nominal aimpoint. 
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    exp_time = parse_value(exp_time, "s")
    roll_angle = parse_value(roll_angle, "deg")
    prng = parse_prng(prng)
    if source is None:
        source_list = []
    elif isinstance(source, dict):
        parameters = {}
        for key in ["flux", "emin", "emax", "src_names"]:
            parameters[key] = source[key]
        source_list = []
        for i in range(len(parameters["flux"])):
            phlist = SimputPhotonList(source["ra"][i], source["dec"][i],
                                      source["energy"][i], parameters['flux'][i],
                                      parameters['src_names'][i])
            source_list.append(phlist)
    elif isinstance(source, str):
        # Assume this is a SIMPUT catalog
        source_list, parameters = read_simput_catalog(source)

    try:
        instrument_spec = instrument_registry[instrument]
    except KeyError:
        raise KeyError(f"Instrument {instrument} is not in the instrument registry!")
    if not instrument_spec["imaging"]:
        raise RuntimeError(f"Instrument '{instrument_spec['name']}' is not "
                           f"designed for imaging observations!")

    arf_file = get_data_file(instrument_spec["arf"])
    rmf_file = get_data_file(instrument_spec["rmf"])
    arf = AuxiliaryResponseFile(arf_file)
    rmf = RedistributionMatrixFile(rmf_file)

    nx = instrument_spec["num_pixels"]
    plate_scale = instrument_spec["fov"]/nx/60. # arcmin to deg
    plate_scale_arcsec = plate_scale * 3600.0

    if aimpt_shift is None:
        aimpt_shift = np.zeros(2)
    aimpt_shift = ensure_numpy_array(aimpt_shift).astype('float64')
    aimpt_shift /= plate_scale_arcsec

    if not instrument_spec["dither"]:
        dither_on = False
    else:
        dither_on = not no_dither
    if dither_params is None:
        dither_params = [8.0, 8.0, 1000.0, 707.0]
    dither_dict = {"x_amp": dither_params[0],
                   "y_amp": dither_params[1],
                   "x_period": dither_params[2],
                   "y_period": dither_params[3],
                   "dither_on": dither_on,
                   "plate_scale": plate_scale_arcsec}

    event_params = {"exposure_time": exp_time,
                    "arf": arf.filename,
                    "sky_center": sky_center,
                    "pix_center": np.array([0.5*(2*nx+1)]*2),
                    "num_pixels": nx,
                    "plate_scale": plate_scale,
                    "rmf": rmf.filename,
                    "channel_type": rmf.chan_type.upper(),
                    "telescope": rmf.header["TELESCOP"],
                    "instrument": instrument_spec['name'],
                    "mission": rmf.header.get("MISSION", ""),
                    "nchan": rmf.n_ch,
                    "roll_angle": roll_angle,
                    "fov": instrument_spec["fov"],
                    "chan_lim": [rmf.cmin, rmf.cmax],
                    "chips": instrument_spec["chips"],
                    "dither_params": dither_dict,
                    "aimpt_coords": instrument_spec["aimpt_coords"],
                    "aimpt_shift": aimpt_shift}

    # Set up WCS

    w = pywcs.WCS(naxis=2)
    w.wcs.crval = event_params["sky_center"]
    w.wcs.crpix = event_params["pix_center"]
    w.wcs.cdelt = [-plate_scale, plate_scale]
    w.wcs.ctype = ["RA---TAN","DEC--TAN"]
    w.wcs.cunit = ["deg"]*2

    # Determine rotation matrix
    rot_mat = get_rot_mat(roll_angle)

    # Set up PSF
    psf_type = instrument_spec["psf"][0]
    psf_class = psf_model_registry[psf_type]
    psf = psf_class(instrument_spec, prng=prng)

    all_events = defaultdict(list)

    for i, src in enumerate(source_list):

        mylog.info(f"Detecting events from source {parameters['src_names'][i]}")

        # Step 1: Use ARF to determine which photons are observed

        mylog.info(f"Applying energy-dependent effective area from "
                   f"{os.path.split(arf.filename)[-1]}.")
        refband = [parameters["emin"][i], parameters["emax"][i]]
        if src.src_type == "phlist":
            events = arf.detect_events_phlist(src.events.copy(), exp_time,
                                              parameters["flux"][i], 
                                              refband, prng=prng)
        elif src.src_type.endswith("spectrum"):
            events = arf.detect_events_spec(src, exp_time, refband, prng=prng)

        n_evt = events["energy"].size

        if n_evt == 0:
            mylog.warning("No events were observed for this source!!!")
        else:

            # Step 2: Assign pixel coordinates to events. Apply dithering and
            # PSF. Clip events that don't fall within the detection region.

            mylog.info("Pixeling events.")

            # Convert RA, Dec to pixel coordinates
            xpix, ypix = w.wcs_world2pix(events["ra"], events["dec"], 1)

            xpix -= event_params["pix_center"][0]
            ypix -= event_params["pix_center"][1]

            events.pop("ra")
            events.pop("dec")

            n_evt = xpix.size

            # Rotate physical coordinates to detector coordinates

            det = np.dot(rot_mat, np.array([xpix, ypix]))
            detx = det[0, :] + event_params["aimpt_coords"][0] + aimpt_shift[0]
            dety = det[1, :] + event_params["aimpt_coords"][1] + aimpt_shift[1]

            # Add times to events
            events['time'] = prng.uniform(size=n_evt, low=0.0,
                                          high=event_params["exposure_time"])

            # Apply dithering

            x_offset, y_offset = perform_dither(events["time"], dither_dict)

            detx -= x_offset
            dety -= y_offset

            # PSF scattering of detector coordinates

            mylog.info(f"Scattering events with a {psf}-based PSF.")
            detx, dety = psf.scatter(detx, dety, events["energy"])

            # Convert detector coordinates to chip coordinates.
            # Throw out events that don't fall on any chip.

            cx = np.trunc(detx)+0.5*np.sign(detx)
            cy = np.trunc(dety)+0.5*np.sign(dety)

            events["chip_id"] = -np.ones(n_evt, dtype='int')
            for i, chip in enumerate(event_params["chips"]):
                rtype = chip[0]
                args = chip[1:]
                r, _ = create_region(rtype, args, 0.0, 0.0)
                inside = r.contains(PixCoord(cx, cy))
                events["chip_id"][inside] = i
            keep = events["chip_id"] > -1

            mylog.info(f"{n_evt-keep.sum()} events were rejected because "
                       f"they do not fall on any CCD.")
            n_evt = keep.sum()

            if n_evt == 0:
                mylog.warning("No events are within the field "
                              "of view for this source!!!")
            else:

                # Keep only those events which fall on a chip

                for key in events:
                    events[key] = events[key][keep]

                # Convert chip coordinates back to detector coordinates,
                # unless the user has specified that they want subpixel
                # resolution

                if subpixel_res:
                    events["detx"] = detx[keep]
                    events["dety"] = dety[keep]
                else:
                    events["detx"] = cx[keep] + \
                                     prng.uniform(low=-0.5, high=0.5, size=n_evt)
                    events["dety"] = cy[keep] + \
                                     prng.uniform(low=-0.5, high=0.5, size=n_evt)

                # Convert detector coordinates back to pixel coordinates by
                # adding the dither offsets back in and applying the rotation
                # matrix again

                det = np.array([events["detx"] + x_offset[keep] -
                                event_params["aimpt_coords"][0] -
                                aimpt_shift[0],
                                events["dety"] + y_offset[keep] -
                                event_params["aimpt_coords"][1] -
                                aimpt_shift[1]])
                pix = np.dot(rot_mat.T, det)

                events["xpix"] = pix[0,:] + event_params['pix_center'][0]
                events["ypix"] = pix[1,:] + event_params['pix_center'][1]

        if n_evt > 0:
            for key in events:
                all_events[key] = np.concatenate([all_events[key], events[key]])

    if len(all_events["energy"]) == 0:
        mylog.warning("No events from any of the sources in "
                      "the catalog were detected!")
        for key in ["xpix", "ypix", "detx", "dety", "time", 
                    "chip_id", event_params["channel_type"]]:
            all_events[key] = np.array([])
    else:
        # Step 4: Scatter energies with RMF
        mylog.info(f"Scattering energies with "
                   f"RMF {os.path.split(rmf.filename)[-1]}.")
        all_events = rmf.scatter_energies(all_events, prng=prng)

    return all_events, event_params


def make_background(exp_time, instrument, sky_center, foreground=True,
                    ptsrc_bkgnd=True, instr_bkgnd=True, no_dither=False,
                    dither_params=None, roll_angle=0.0, subpixel_res=False,
                    input_pt_sources=None, aimpt_shift=None, prng=None,
                    **kwargs):
    """
    Make background events.

    Parameters
    ----------
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time to use, in seconds. 
    instrument : string
        The name of the instrument to use, which picks an instrument
        specification from the instrument registry. 
    sky_center : array, tuple, or list
        The center RA, Dec coordinates of the observation, in degrees.
    foreground : boolean, optional
        Whether or not to include the Galactic foreground. Default: True
    instr_bkgnd : boolean, optional
        Whether or not to include the instrumental background. Default: True
    no_dither : boolean, optional
        If True, turn off dithering entirely. Default: False
    dither_params : array-like of floats, optional
        The parameters to use to control the size and period of the dither
        pattern. The first two numbers are the dither amplitude in x and y
        detector coordinates in arcseconds, and the second two numbers are
        the dither period in x and y detector coordinates in seconds. 
        Default: [8.0, 8.0, 1000.0, 707.0].
    ptsrc_bkgnd : boolean, optional
        Whether or not to include the point-source background. Default: True
    roll_angle : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The roll angle of the observation in degrees. Default: 0.0
    subpixel_res: boolean, optional
        If True, event positions are not randomized within the pixels 
        within which they are detected. Default: False
    input_pt_sources : string, optional
        If set to a filename, input the point source positions, fluxes,
        and spectral indices from an ASCII table instead of generating
        them. Default: None
    aimpt_shift : array-like, optional
        A two-float array-like object which shifts the aimpoint on the 
        detector from the nominal position. Units are in arcseconds.
        Default: None, which results in no shift from the nominal aimpoint. 
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    if "nH" in kwargs or "absorb_model" in kwargs:
        warnings.warn("The 'nH' and 'absorb_model' keyword arguments"
                      "have been omitted. Please set the 'bkgnd_nH' "
                      "and 'bkgnd_absorb_model' values in the SOXS"
                      "configuration file if you want to change these "
                      "values. ",
                      DeprecationWarning)
    from soxs.background import make_instrument_background, \
        make_foreground, make_ptsrc_background
    prng = parse_prng(prng)
    exp_time = parse_value(exp_time, "s")
    roll_angle = parse_value(roll_angle, "deg")
    try:
        instrument_spec = instrument_registry[instrument]
    except KeyError:
        raise KeyError(f"Instrument {instrument} is not in the "
                       f"instrument registry!")
    if not instrument_spec["imaging"]:
        raise RuntimeError(f"Instrument '{instrument_spec['name']}' is not "
                           f"designed for imaging observations!")
    fov = instrument_spec["fov"]

    input_events = defaultdict(list)

    arf_file = get_data_file(instrument_spec["arf"])
    arf = AuxiliaryResponseFile(arf_file)
    rmf_file = get_data_file(instrument_spec["rmf"])
    rmf = RedistributionMatrixFile(rmf_file)

    if ptsrc_bkgnd:
        mylog.info("Adding in point-source background.")
        ptsrc_events = make_ptsrc_background(exp_time, fov, sky_center,
                                             area=1.2*arf.max_area,
                                             input_sources=input_pt_sources,
                                             prng=prng)
        for key in ["ra", "dec", "energy"]:
            input_events[key].append(ptsrc_events[key])
        input_events["flux"].append(ptsrc_events["flux"])
        input_events["emin"].append(ptsrc_events["energy"].min())
        input_events["emax"].append(ptsrc_events["energy"].max())
        input_events["src_names"].append("ptsrc_bkgnd")
        events, event_params = generate_events(input_events, exp_time,
                                               instrument, sky_center,
                                               no_dither=no_dither,
                                               dither_params=dither_params,
                                               roll_angle=roll_angle,
                                               subpixel_res=subpixel_res,
                                               aimpt_shift=aimpt_shift,
                                               prng=prng)
        mylog.info(f"Generated {events['energy'].size} photons from "
                   f"the point-source background.")
    else:
        nx = instrument_spec["num_pixels"]
        plate_scale = instrument_spec["fov"]/nx/60.0
        plate_scale_arcsec = plate_scale*3600.0
        if aimpt_shift is None:
            aimpt_shift = np.zeros(2)
        aimpt_shift = ensure_numpy_array(aimpt_shift).astype('float64')
        aimpt_shift /= plate_scale_arcsec
        events = defaultdict(list)
        if not instrument_spec["dither"]:
            dither_on = False
        else:
            dither_on = not no_dither
        if dither_params is None:
            dither_params = [8.0, 8.0, 1000.0, 707.0]
        dither_dict = {"x_amp": dither_params[0],
                       "y_amp": dither_params[1],
                       "x_period": dither_params[2],
                       "y_period": dither_params[3],
                       "dither_on": dither_on,
                       "plate_scale": instrument_spec["fov"]/nx*60.0}
        event_params = {"exposure_time": exp_time, 
                        "fov": instrument_spec["fov"],
                        "num_pixels": nx,
                        "pix_center": np.array([0.5*(2*nx+1)]*2),
                        "channel_type": rmf.header["CHANTYPE"].upper(),
                        "sky_center": sky_center,
                        "dither_params": dither_dict,
                        "plate_scale": plate_scale,
                        "chan_lim": [rmf.cmin, rmf.cmax],
                        "rmf": rmf_file, "arf": arf_file,
                        "telescope": rmf.header["TELESCOP"],
                        "instrument": instrument_spec['name'],
                        "mission": rmf.header.get("MISSION", ""),
                        "nchan": rmf.n_ch,
                        "roll_angle": roll_angle,
                        "aimpt_coords": instrument_spec["aimpt_coords"],
                        "aimpt_shift": aimpt_shift}

    if "chips" not in event_params:
        event_params["chips"] = instrument_spec["chips"]

    if foreground:
        mylog.info("Adding in astrophysical foreground.")
        bkg_events = make_foreground(event_params, arf, rmf, prng=prng)
        for key in bkg_events:
            events[key] = np.concatenate([events[key], bkg_events[key]])
    if instr_bkgnd and instrument_spec["bkgnd"] is not None:
        mylog.info("Adding in instrumental background.")
        bkg_events = make_instrument_background(instrument_spec,
                                                event_params, rmf, prng=prng)
        for key in bkg_events:
            events[key] = np.concatenate([events[key], bkg_events[key]])

    return events, event_params


def make_background_file(out_file, exp_time, instrument, sky_center,
                         overwrite=False, foreground=True, instr_bkgnd=True,
                         ptsrc_bkgnd=True, no_dither=False, dither_params=None,
                         subpixel_res=False, input_pt_sources=None, 
                         prng=None, **kwargs):
    """
    Make an event file consisting entirely of background events. This will be 
    useful for creating backgrounds that can be added to simulations of sources.

    Parameters
    ----------
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time to use, in seconds. 
    instrument : string
        The name of the instrument to use, which picks an instrument
        specification from the instrument registry. 
    sky_center : array, tuple, or list
        The center RA, Dec coordinates of the observation, in degrees.
    overwrite : boolean, optional
        Whether or not to overwrite an existing file with the same name.
        Default: False
    foreground : boolean, optional
        Whether or not to include the Galactic foreground. Default: True
    instr_bkgnd : boolean, optional
        Whether or not to include the instrumental background. Default: True
    ptsrc_bkgnd : boolean, optional
        Whether or not to include the point-source background. Default: True
    no_dither : boolean, optional
        If True, turn off dithering entirely. Default: False
    dither_params : array-like of floats, optional
        The parameters to use to control the size and period of the dither
        pattern. The first two numbers are the dither amplitude in x and y
        detector coordinates in arcseconds, and the second two numbers are
        the dither period in x and y detector coordinates in seconds. 
        Default: [8.0, 8.0, 1000.0, 707.0].
    subpixel_res: boolean, optional
        If True, event positions are not randomized within the pixels 
        within which they are detected. Default: False
    input_pt_sources : string, optional
        If set to a filename, input the point source positions, fluxes,
        and spectral indices from an ASCII table instead of generating
        them. Default: None
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    if "nH" in kwargs or "absorb_model" in kwargs:
        warnings.warn("The 'nH' and 'absorb_model' keyword arguments"
                      "have been omitted. Please set the 'bkgnd_nH' "
                      "and 'bkgnd_absorb_model' values in the SOXS"
                      "configuration file if you want to change these "
                      "values. ",
                      DeprecationWarning)
    if "input_sources" in kwargs:
        warnings.warn("The 'input_sources' keyword argument has been changed "
                      "to 'input_pt_sources' and is deprecated.", 
                      DeprecationWarning)
        input_pt_sources = kwargs.pop("input_sources")
    prng = parse_prng(prng)
    events, event_params = make_background(exp_time, instrument, sky_center, 
                                           ptsrc_bkgnd=ptsrc_bkgnd, 
                                           foreground=foreground, 
                                           instr_bkgnd=instr_bkgnd,
                                           no_dither=no_dither,
                                           dither_params=dither_params, 
                                           subpixel_res=subpixel_res,
                                           input_pt_sources=input_pt_sources,
                                           prng=prng)
    write_event_file(events, event_params, out_file, overwrite=overwrite)


def instrument_simulator(input_events, out_file, exp_time, instrument,
                         sky_center, overwrite=False, instr_bkgnd=True, 
                         foreground=True, ptsrc_bkgnd=True, 
                         bkgnd_file=None, no_dither=False, 
                         dither_params=None, roll_angle=0.0, 
                         subpixel_res=False, aimpt_shift=None,
                         input_pt_sources=None, prng=None):
    """
    Take unconvolved events and create an event file from them. This
    function calls generate_events to do the following:

    1. Determines which events are observed using the ARF
    2. Pixelizes the events, applying PSF effects and dithering
    3. Determines energy channels using the RMF

    and then calls make_background to add instrumental and astrophysical
    backgrounds, unless a background file is provided, in which case
    the background events are read from this file. The events are
    then written out to a file.

    Parameters
    ----------
    input_events : string, dict, or None
        The unconvolved events to be used as input. Can be one of the
        following:
        1. The name of a SIMPUT catalog file.
        2. A Python dictionary containing the following items:
        "ra": A NumPy array of right ascension values in degrees.
        "dec": A NumPy array of declination values in degrees.
        "energy": A NumPy array of energy values in keV.
        "flux": The flux of the entire source, in units of erg/cm**2/s.
    out_file : string
        The name of the event file to be written.
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time to use, in seconds. 
    instrument : string
        The name of the instrument to use, which picks an instrument
        specification from the instrument registry. 
    sky_center : array, tuple, or list
        The center RA, Dec coordinates of the observation, in degrees.
    overwrite : boolean, optional
        Whether or not to overwrite an existing file with the same name.
        Default: False
    instr_bkgnd : boolean, optional
        Whether or not to include the instrumental/particle background. 
        Default: True
    foreground : boolean, optional
        Whether or not to include the local foreground. 
        Default: True
    ptsrc_bkgnd : boolean, optional
        Whether or not to include the point-source background. 
        Default: True
    bkgnd_file : string, optional
        If set, backgrounds will be loaded from this file and not generated
        on the fly. Default: None
    no_dither : boolean, optional
        If True, turn off dithering entirely. Default: False
    dither_params : array-like of floats, optional
        The parameters to use to control the size and period of the dither
        pattern. The first two numbers are the dither amplitude in x and y
        detector coordinates in arcseconds, and the second two numbers are
        the dither period in x and y detector coordinates in seconds. 
        Default: [8.0, 8.0, 1000.0, 707.0].
    roll_angle : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The roll angle of the observation in degrees. Default: 0.0
    subpixel_res: boolean, optional
        If True, event positions are not randomized within the pixels 
        within which they are detected. Default: False
    aimpt_shift : array-like, optional
        A two-float array-like object which shifts the aimpoint on the 
        detector from the nominal position. Units are in arcseconds.
        Default: None, which results in no shift from the nominal aimpoint. 
    input_pt_sources : string, optional
        If set to a filename, input the point source positions, fluxes,
        and spectral indices from an ASCII table instead of generating
        them. Default: None
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 

    Examples
    --------
    >>> instrument_simulator("sloshing_simput.fits", "sloshing_evt.fits", 
    ...                      300000.0, "lynx_hdxi", [30., 45.], overwrite=True)
    """
    from soxs.background import add_background_from_file
    if not out_file.endswith(".fits"):
        out_file += ".fits"
    mylog.info(f"Making observation of source in {out_file}.")
    # Make the source first
    events, event_params = generate_events(input_events, exp_time, instrument, sky_center,
                                           no_dither=no_dither, dither_params=dither_params, 
                                           roll_angle=roll_angle, subpixel_res=subpixel_res, 
                                           aimpt_shift=aimpt_shift, prng=prng)
    # If the user wants backgrounds, either make the background or add an already existing
    # background event file. It may be necessary to reproject events to a new coordinate system.
    if bkgnd_file is None:
        if not instr_bkgnd and not ptsrc_bkgnd and not foreground:
            mylog.info("No backgrounds will be added to this observation.")
        else:
            mylog.info("Adding background events.")
            bkg_events, _ = make_background(
                exp_time, instrument, sky_center, foreground=foreground,
                instr_bkgnd=instr_bkgnd, no_dither=no_dither,
                dither_params=dither_params, ptsrc_bkgnd=ptsrc_bkgnd, prng=prng,
                subpixel_res=subpixel_res, roll_angle=roll_angle,
                aimpt_shift=aimpt_shift, input_pt_sources=input_pt_sources)
            for key in events:
                events[key] = np.concatenate([events[key], bkg_events[key]])
    else:
        mylog.info(f"Adding background events from the file {bkgnd_file}.")
        if not os.path.exists(bkgnd_file):
            raise IOError(f"Cannot find the background event file {bkgnd_file}!")
        events = add_background_from_file(events, event_params, bkgnd_file)
    if len(events["energy"]) == 0:
        mylog.warning("No events were detected from source or background!! We "
                      "will not write an event file.")
    else:
        write_event_file(events, event_params, out_file, overwrite=overwrite)
    mylog.info("Observation complete.")


def simulate_spectrum(spec, instrument, exp_time, out_file,
                      instr_bkgnd=False, foreground=False,
                      ptsrc_bkgnd=False, bkgnd_area=None,
                      overwrite=False, prng=None, **kwargs):
    """
    Generate a PI or PHA spectrum from a :class:`~soxs.spectra.Spectrum`
    by convolving it with responses. To be used if one wants to 
    create a spectrum without worrying about spatial response. Similar
    to XSPEC's "fakeit".

    Parameters
    ----------
    spec : :class:`~soxs.spectra.Spectrum`
        The spectrum to be convolved. If None is supplied, only backgrounds
        will be simulated (if they are turned on).
    instrument : string
        The name of the instrument to use, which picks an instrument
        specification from the instrument registry.
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time in seconds.
    out_file : string
        The file to write the spectrum to.
    instr_bkgnd : boolean, optional
        Whether or not to include the instrumental/particle background. 
        Default: False
    foreground : boolean, optional
        Whether or not to include the local foreground.
        Default: False
    ptsrc_bkgnd : boolean, optional
        Whether or not to include the unresolved point-source background. 
        Default: False
    bkgnd_area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The area on the sky for the background components, in square arcminutes.
        Default: None, necessary to specify if any of the background components
        are turned on. 
    overwrite : boolean, optional
        Whether or not to overwrite an existing file. Default: False
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 

    Examples
    --------
    >>> spec = soxs.Spectrum.from_file("my_spectrum.txt")
    >>> soxs.simulate_spectrum(spec, "lynx_lxm", 100000.0, 
    ...                        "my_spec.pi", overwrite=True)
    """
    from soxs.events import _write_spectrum
    from soxs.response import RedistributionMatrixFile, \
        AuxiliaryResponseFile
    from soxs.spectra import ConvolvedSpectrum
    from soxs.background.foreground import make_frgnd_spectrum
    from soxs.background.spectra import BackgroundSpectrum
    from soxs.background.instrument import InstrumentalBackground
    from soxs.utils import soxs_cfg
    if "nH" in kwargs or "absorb_model" in kwargs:
        warnings.warn("The 'nH' and 'absorb_model' keyword arguments"
                      "have been omitted. Please set the 'bkgnd_nH' "
                      "and 'bkgnd_absorb_model' values in the SOXS"
                      "configuration file if you want to change these "
                      "values. ",
                      DeprecationWarning)
    prng = parse_prng(prng)
    exp_time = parse_value(exp_time, "s")
    try:
        instrument_spec = instrument_registry[instrument]
    except KeyError:
        raise KeyError(f"Instrument {instrument} is not in the instrument registry!")
    if foreground or instr_bkgnd or ptsrc_bkgnd:
        if instrument_spec["grating"]:
            raise NotImplementedError("Backgrounds cannot be included in simulations "
                                      "of gratings spectra at this time!")
        if bkgnd_area is None:
            raise RuntimeError("The 'bkgnd_area' argument must be set if one wants "
                               "to simulate backgrounds! Specify a value in square "
                               "arcminutes.")
        bkgnd_area = np.sqrt(parse_value(bkgnd_area, "arcmin**2"))
    elif spec is None:
        raise RuntimeError("You have specified no source spectrum and no backgrounds!")
    arf_file = get_data_file(instrument_spec["arf"])
    rmf_file = get_data_file(instrument_spec["rmf"])
    arf = AuxiliaryResponseFile(arf_file)
    rmf = RedistributionMatrixFile(rmf_file)

    event_params = {"RESPFILE": os.path.split(rmf.filename)[-1],
                    "ANCRFILE": os.path.split(arf.filename)[-1],
                    "TELESCOP": rmf.header["TELESCOP"],
                    "INSTRUME": rmf.header["INSTRUME"],
                    "MISSION": rmf.header.get("MISSION", "")}

    out_spec = np.zeros(rmf.n_ch)

    if spec is not None:
        cspec = ConvolvedSpectrum.convolve(spec, arf)
        out_spec += rmf.convolve_spectrum(cspec, exp_time, prng=prng)

    fov = None if bkgnd_area is None else np.sqrt(bkgnd_area)

    if foreground:
        mylog.info("Adding in astrophysical foreground.")
        cspec_frgnd = ConvolvedSpectrum.convolve(
            make_frgnd_spectrum.spec.to_spectrum(fov), arf)
        out_spec += rmf.convolve_spectrum(cspec_frgnd, exp_time, prng=prng)
    if instr_bkgnd and instrument_spec["bkgnd"] is not None:
        mylog.info("Adding in instrumental background.")
        bkgnd_spec = instrument_spec["bkgnd"]
        # Temporary hack for ACIS-S
        if "aciss" in instrument_spec["name"]:
            bkgnd_spec = bkgnd_spec[1]
        bkgnd_spec = InstrumentalBackground.from_filename(
            bkgnd_spec[0], bkgnd_spec[1],
            instrument_spec['focal_length'])
        out_spec += bkgnd_spec.generate_channel_spectrum(exp_time, bkgnd_area,
                                                         prng=prng)
    if ptsrc_bkgnd:
        mylog.info("Adding in background from unresolved point-sources.")
        bkgnd_nH = float(soxs_cfg.get("soxs", "bkgnd_nH"))
        absorb_model = soxs_cfg.get("soxs", "bkgnd_absorb_model")
        spec_plaw = BackgroundSpectrum.from_powerlaw(1.52, 0.0, 2.0e-7, emin=0.01,
                                                     emax=10.0, nbins=300000)
        spec_plaw.apply_foreground_absorption(bkgnd_nH, model=absorb_model)
        cspec_plaw = ConvolvedSpectrum.convolve(spec_plaw.to_spectrum(fov), arf)
        out_spec += rmf.convolve_spectrum(cspec_plaw, exp_time, prng=prng)

    bins = (np.arange(rmf.n_ch)+rmf.cmin).astype("int32")

    _write_spectrum(bins, out_spec, exp_time, rmf.header["CHANTYPE"], 
                    event_params, out_file, overwrite=overwrite)



from soxs.background import \
    BackgroundSpectrum, \
    ConvolvedBackgroundSpectrum, \
    make_point_sources_file, \
    make_point_source_list, \
    InstrumentalBackground

from soxs.cosmology import \
    make_cosmological_sources_file

from soxs.events import \
    write_spectrum, \
    write_image, \
    write_radial_profile, \
    plot_spectrum, \
    make_exposure_map, \
    plot_image, \
    filter_events

from soxs.instrument import \
    instrument_simulator, \
    make_background_file, \
    simulate_spectrum

from soxs.instrument_registry import \
    add_instrument_to_registry, \
    show_instrument_registry, \
    write_instrument_json, \
    get_instrument_from_registry, \
    instrument_registry, \
    make_simple_instrument

from soxs.mosaic import \
    make_mosaic_events, \
    make_mosaic_image

from soxs.response import \
    AuxiliaryResponseFile, \
    RedistributionMatrixFile, \
    FlatResponse

from soxs.simput import \
    read_simput_catalog, \
    SimputPhotonList, \
    SimputCatalog, \
    SimputSpectrum, \
    write_photon_list

from soxs.spatial import \
    PointSourceModel, \
    RadialFunctionModel, \
    RadialArrayModel, \
    RadialFileModel, \
    AnnulusModel, \
    BetaModel, \
    DoubleBetaModel, \
    FillFOVModel, \
    RectangleModel, \
    SpatialModel

from soxs.spectra import \
    Spectrum, \
    ApecGenerator, \
    ConvolvedSpectrum

from soxs.utils import \
    set_soxs_config, \
    set_mission_config

__version__ = "3.3.0"



from soxs.instrument import make_background, \
    instrument_simulator, make_background_file, simulate_spectrum
from soxs.background.foreground import make_frgnd_spectrum
from soxs.background.spectra import ConvolvedBackgroundSpectrum
from soxs.spectra import Spectrum
from soxs.response import AuxiliaryResponseFile, RedistributionMatrixFile
from soxs.utils import soxs_files_path, set_soxs_config
from soxs.tests.utils import spectrum_answer_testing
from numpy.random import RandomState
from numpy.testing import assert_allclose
import astropy.io.fits as pyfits
import tempfile
import os
import shutil
import numpy as np
import astropy.units as u

acisi_particle_bkgnd = Spectrum.from_file(
    os.path.join(soxs_files_path, "acisi_particle_bkgnd.h5"))


def test_uniform_bkgnd_scale():
    prng = RandomState(25)
    hdxi_arf = AuxiliaryResponseFile("xrs_hdxi_3x10.arf")
    events, event_params = make_background((50, "ks"), "lynx_hdxi", [30., 45.], 
                                           foreground=True, instr_bkgnd=True,
                                           ptsrc_bkgnd=False, prng=prng)
    ncts = np.logical_and(events["energy"] >= 0.7, events["energy"] <= 2.0).sum()
    t_exp = event_params["exposure_time"]
    fov = (event_params["fov"]*60.0)**2
    S = ncts/t_exp/fov
    dS = np.sqrt(ncts)/t_exp/fov
    foreground = ConvolvedBackgroundSpectrum.convolve(make_frgnd_spectrum.spec, hdxi_arf)
    f_sum = foreground.get_flux_in_band(0.7, 2.0)[0]
    i_sum = acisi_particle_bkgnd.get_flux_in_band(0.7, 2.0)[0]*(u.cm/u.arcmin)**2
    b_sum = (f_sum+i_sum).to("ph/(arcsec**2*s)").value
    assert np.abs(S-b_sum) < 1.645*dS


def test_simulate_bkgnd_spectrum():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    prng = RandomState(29)

    hdxi_arf = AuxiliaryResponseFile("xrs_hdxi_3x10.arf")
    hdxi_rmf = RedistributionMatrixFile("xrs_hdxi.rmf")

    exp_time = 50000.0
    fov = 3600.0
    simulate_spectrum(None, "lynx_hdxi", exp_time, "test_bkgnd.pha",
                      instr_bkgnd=True, foreground=True, prng=prng,
                      overwrite=True, bkgnd_area=(fov, "arcsec**2"))
    ch_min = hdxi_rmf.eb_to_ch(0.7)-hdxi_rmf.cmin
    ch_max = hdxi_rmf.eb_to_ch(2.0)-hdxi_rmf.cmin
    with pyfits.open("test_bkgnd.pha") as f:
        ncts = f["SPECTRUM"].data["COUNTS"][ch_min:ch_max].sum()
    S = ncts/exp_time/fov
    dS = np.sqrt(ncts)/exp_time/fov
    foreground = ConvolvedBackgroundSpectrum.convolve(make_frgnd_spectrum.spec, hdxi_arf)
    f_sum = foreground.get_flux_in_band(0.7, 2.0)[0]
    i_sum = acisi_particle_bkgnd.get_flux_in_band(0.7, 2.0)[0]*(u.cm/u.arcmin)**2
    b_sum = (f_sum+i_sum).to_value("ph/(arcsec**2*s)")
    assert np.abs(S-b_sum) < 1.645*dS

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_add_background():
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    prng1 = RandomState(29)
    prng2 = RandomState(29)

    ra0 = 30.0
    dec0 = 45.0
    ra1 = 22.0
    dec1 = 22.0
    exp_time = 50000.0

    instrument_simulator(None, "evt1.fits", exp_time, "lynx_hdxi",
                         [ra0, dec0], prng=prng1, overwrite=True)

    make_background_file("bkg_evt.fits", exp_time, "lynx_hdxi", [ra0, dec0],
                         prng=prng2, overwrite=True)

    instrument_simulator(None, "evt2.fits", exp_time, "lynx_hdxi",
                         [ra1, dec1], bkgnd_file="bkg_evt.fits",
                         prng=prng2, overwrite=True)

    f1 = pyfits.open("evt1.fits")
    f2 = pyfits.open("evt2.fits")

    for key in ["X", "Y", "ENERGY", "PHA"]:
        assert_allclose(f1["EVENTS"].data[key], f2["EVENTS"].data[key], rtol=1.0e-6)

    f1.close()
    f2.close()

    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_ptsrc():
    from soxs.background.point_sources import generate_fluxes, \
        make_ptsrc_background
    from soxs.data import cdf_fluxes, cdf_gal, cdf_agn
    from soxs.constants import erg_per_keV
    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    prng = RandomState(33)
    fov = 20.0
    f_agn = np.zeros((cdf_fluxes.size-1, 100))
    f_gal = np.zeros((cdf_fluxes.size-1, 100))
    for k in range(100):
        agn_fluxes, gal_fluxes = generate_fluxes(fov, prng)
        f_agn[:,k] = np.histogram(agn_fluxes, bins=cdf_fluxes)[0]
        f_gal[:,k] = np.histogram(gal_fluxes, bins=cdf_fluxes)[0]
    mu_agn = np.mean(f_agn, axis=1)
    sigma_agn = np.std(f_agn, axis=1)
    mu_gal = np.mean(f_gal, axis=1)
    sigma_gal = np.std(f_gal, axis=1)
    f_agn0 = np.diff(cdf_agn)*(fov/60.0)**2
    f_gal0 = np.diff(cdf_gal)*(fov/60.0)**2
    err_agn = np.abs(mu_agn-f_agn0)/sigma_agn
    err_gal = np.abs(mu_gal-f_gal0)/sigma_gal
    err_agn[sigma_agn == 0.0] = 0.0
    err_gal[sigma_gal == 0.0] = 0.0
    assert np.all(err_agn < 1.0)
    assert np.all(err_gal < 1.0)
    exp_time = 500000.0 # seconds
    fov = 20.0 # arcmin
    area = 30000.0 # cm**2
    sky_center = [20., 17.]
    prng1 = RandomState(33)
    prng2 = RandomState(33)
    agn_fluxes, gal_fluxes = generate_fluxes(fov, prng2)
    fluxes = np.concatenate([agn_fluxes, gal_fluxes])
    events = make_ptsrc_background(exp_time, fov, sky_center, area=area, 
                                   prng=prng1, nH=-1.0, output_sources="src.dat")
    idxs = np.logical_and(events["energy"] > 0.5, events["energy"] < 2.0)
    E_mean = events["energy"][idxs].mean()*erg_per_keV
    n1 = fluxes.sum()*exp_time*area/E_mean
    n2 = idxs.sum()
    dn = np.sqrt(n2)
    assert np.abs(n1-n2) < 1.645*dn
    events2 = make_ptsrc_background(exp_time, fov, sky_center, area=area,
                                    prng=prng1, nH=-1.0, input_sources="src.dat")
    assert_allclose(events["ra"].sum(), events2["ra"].sum(), rtol=1.0e-3)
    assert_allclose(events["dec"].sum(), events2["dec"].sum(), rtol=1.0e-3)
    assert_allclose(events["energy"].sum(), events2["energy"].sum(), rtol=1.0e-3)
    os.chdir(curdir)
    shutil.rmtree(tmpdir)


def test_change_bkgnd(answer_store, answer_dir):
    from soxs.background.foreground import make_frgnd_spectrum
    set_soxs_config("frgnd_spec_model", "default")
    spectrum_answer_testing(make_frgnd_spectrum.spec, 
                            f"default_frgnd_spectrum.h5", answer_store,
                            answer_dir)
    set_soxs_config("frgnd_spec_model", "halosat")
    spectrum_answer_testing(make_frgnd_spectrum.spec,
                            f"lem_frgnd_spectrum.h5", answer_store,
                            answer_dir)


if __name__ == "__main__":
    test_add_background()
    test_uniform_bkgnd_scale()
    test_ptsrc()
    test_simulate_bkgnd_spectrum()


import os
import logging
import numpy as np
from numpy.random import RandomState
import astropy.units as u
from astropy.units import Quantity
import warnings
from configparser import ConfigParser
import regions
import appdirs

# Configuration

soxs_cfg_defaults = {"soxs_data_dir": "/does/not/exist",
                     "abund_table": "angr",
                     "apec_vers": "3.0.9",
                     "bkgnd_nH": 0.018,
                     "bkgnd_absorb_model": "wabs",
                     "frgnd_spec_model": "default"}

CONFIG_DIR = os.environ.get('XDG_CONFIG_HOME',
                            os.path.join(os.path.expanduser('~'),
                                         '.config', 'soxs'))
if not os.path.exists(CONFIG_DIR):
    try:
        os.makedirs(CONFIG_DIR)
    except OSError:
        warnings.warn("unable to create soxs config directory")

CURRENT_CONFIG_FILE = os.path.join(CONFIG_DIR, 'soxs.cfg')

if not os.path.exists(CURRENT_CONFIG_FILE):
    cp = ConfigParser()
    cp.add_section("soxs")
    try:
        with open(CURRENT_CONFIG_FILE, 'w') as new_cfg:
            cp.write(new_cfg)
    except IOError:
        warnings.warn("unable to write new config file")


soxs_cfg = ConfigParser(soxs_cfg_defaults)
soxs_cfg.read([CURRENT_CONFIG_FILE, 'soxs.cfg'])
if not soxs_cfg.has_section("soxs"):
    soxs_cfg.add_section("soxs")

# Logging

soxsLogger = logging.getLogger("soxs")

ufstring = "%(name)-3s : [%(levelname)-9s] %(asctime)s %(message)s"
cfstring = "%(name)-3s : [%(levelname)-18s] %(asctime)s %(message)s"

soxs_sh = logging.StreamHandler()
# create formatter and add it to the handlers
formatter = logging.Formatter(ufstring)
soxs_sh.setFormatter(formatter)
# add the handler to the logger
soxsLogger.addHandler(soxs_sh)
soxsLogger.setLevel('INFO')
soxsLogger.propagate = False

mylog = soxsLogger

mylog.setLevel('INFO')

if soxs_cfg.has_option("soxs", "response_path"):
    mylog.warning("The 'response_path' option in the SOXS configuration "
                  "is deprecated and has been replaced with 'soxs_data_dir'. "
                  "Please update your configuration accordingly.")
    soxs_cfg.set("soxs", "soxs_data_dir", soxs_cfg.get("soxs", "response_path"))

if soxs_cfg.get("soxs", "soxs_data_dir") == "/does/not/exist":
    soxs_data_dir = appdirs.user_cache_dir("soxs")
    mylog.warning(f"Setting 'soxs_data_dir' to {soxs_data_dir} for this session. "
                  f"Please update your configuration if you want it somewhere else.")
    soxs_cfg.set("soxs", "soxs_data_dir", appdirs.user_cache_dir("soxs"))


def issue_deprecation_warning(msg):
    import warnings
    from numpy import VisibleDeprecationWarning
    warnings.warn(msg, VisibleDeprecationWarning, stacklevel=3)


soxs_path = os.path.abspath(os.path.dirname(__file__))
soxs_files_path = os.path.join(soxs_path, "files")


def parse_prng(prng):
    if isinstance(prng, RandomState):
        return prng
    else:
        return RandomState(prng)


def iterable(obj):
    """
    Grabbed from Python Cookbook / matplotlib.cbook.
    Returns true/false for *obj* iterable.
    """
    try:
        len(obj)
    except:
        return False
    return True


def ensure_list(obj):
    """
    This function ensures that *obj* is a list.  Typically used to convert a
    string to a list, for instance ensuring the *fields* as an argument is a
    list.
    """
    if obj is None:
        return [obj]
    if not isinstance(obj, list):
        return [obj]
    return obj


def ensure_numpy_array(obj):
    """
    This function ensures that *obj* is a numpy array. 
    Typically used to convert scalar, list or tuple 
    argument passed to functions using Cython.
    """
    if isinstance(obj, np.ndarray):
        if obj.shape == ():
            return np.array([obj])
        # We cast to ndarray to catch ndarray subclasses
        return np.array(obj)
    elif isinstance(obj, (list, tuple)):
        return np.asarray(obj)
    else:
        return np.asarray([obj])


def parse_value(value, default_units, equivalence=None):
    if isinstance(value, str):
        v = value.split(",")
        if len(v) == 2:
            value = (float(v[0]), v[1])
        else:
            value = float(v[0])
    if hasattr(value, "to_astropy"):
        value = value.to_astropy()
    if isinstance(value, Quantity):
        q = Quantity(value.value, value.unit)
    elif iterable(value):
        q = Quantity(value[0], value[1])
    else:
        q = Quantity(value, default_units)
    return q.to(default_units, equivalencies=equivalence).value


def get_rot_mat(roll_angle):
    roll_angle = np.deg2rad(roll_angle)
    rot_mat = np.array([[np.cos(roll_angle), -np.sin(roll_angle)],
                        [np.sin(roll_angle), np.cos(roll_angle)]])
    return rot_mat


def downsample(myarr,factor,estimator=np.mean):
    """
    Downsample a 2D array by averaging over *factor* pixels in each axis.
    Crops upper edge if the shape is not a multiple of factor.

    This code is pure numpy and should be fast.

    keywords:
        estimator - default to mean.  You can downsample by summing or
            something else if you want a different estimator
            (e.g., downsampling error: you want to sum & divide by sqrt(n))
    """
    ys,xs = myarr.shape
    crarr = myarr[:ys-(ys % int(factor)),:xs-(xs % int(factor))]
    dsarr = estimator(np.concatenate([[crarr[i::factor,j::factor]
                                       for i in range(factor)]
                                      for j in range(factor)]), axis=0)
    return dsarr


def line_width_equiv(rest):
    from astropy.constants import c
    ckms = c.to_value('km/s')
    forward = lambda x: rest*x/ckms
    backward = lambda x: x/rest*ckms
    return [(u.km/u.s, u.keV, forward, backward)]


class DummyPbar(object):
    def __init__(self):
        pass

    def update(self, *args, **kwargs):
        pass

    def close(self):
        pass


def create_region(rtype, args, dx, dy):
    if rtype in ["Rectangle", "Box"]:
        xctr, yctr, xw, yw = args
        x = xctr+dx
        y = yctr+dy
        center = regions.PixCoord(x=x, y=y)
        reg = regions.RectanglePixelRegion(center=center, width=xw, height=yw)
        bounds = [x-0.5*xw, x+0.5*xw, y-0.5*yw, y+0.5*yw]
    elif rtype == "Circle":
        xctr, yctr, radius = args
        x = xctr+dx
        y = yctr+dy
        center = regions.PixCoord(x=x, y=y)
        reg = regions.CirclePixelRegion(center=center, radius=radius)
        bounds = [x-radius, x+radius, y-radius, y+radius]
    elif rtype == "Polygon":
        x = np.array(args[0])+dx
        y = np.array(args[1])+dy
        vertices = regions.PixCoord(x=x, y=y)
        reg = regions.PolygonPixelRegion(vertices=vertices)
        bounds = [x.min(), x.max(), y.min(), y.max()]
    else:
        raise NotImplementedError
    return reg, bounds


def process_fits_string(fitsstr):
    import re
    import astropy.io.fits as pyfits
    fn = fitsstr.split("[")[0]
    brackets = re.findall(r"[^[]*\[([^]]*)\]", fitsstr)
    with pyfits.open(fn) as f:
        if len(brackets) == 0:
            imgs = np.array([hdu.is_image for hdu in f])
            if imgs.sum() > 1:
                raise IOError("Multiple HDUs in this file, "
                              "please specify one to read!")
            ext = np.where(imgs)[0][0]
        else:
            ext = brackets[0]
            if ext.isdigit():
                ext = int(ext)
        imhdu = f[ext]
    return imhdu


class PoochHandle:
    r"""
    Container for a pooch object used to fetch remote response that isn't
    already stored locally.
    """
    def __init__(self, cache_dir=None):
        import json
        import pooch
        import pkg_resources
        if cache_dir is None:
            if os.path.isdir(soxs_cfg.get("soxs", "soxs_data_dir")):
                cache_dir = soxs_cfg.get("soxs", "soxs_data_dir")
            else:
                cache_dir = pooch.os_cache("soxs")
        self._registry = json.load(
            pkg_resources.resource_stream("soxs", "file_hash_registry.json"))
        self.pooch_obj = pooch.create(
            path=cache_dir,
            registry=self._registry,
            env="SOXS_DATA_DIR",
            base_url="https://hea-www.cfa.harvard.edu/soxs/soxs_responses/"
        )
        self.dl = pooch.HTTPDownloader(progressbar=True)

    def fetch(self, fname):
        return self.pooch_obj.fetch(fname, downloader=self.dl)


finley = PoochHandle()


def get_data_file(fn):
    soxs_data_dir = soxs_cfg.get("soxs", "soxs_data_dir")
    rel_fn = os.path.split(fn)[-1]
    data_fn = os.path.join(soxs_data_dir, rel_fn)
    if os.path.exists(fn):
        return fn
    elif os.path.exists(data_fn):
        return data_fn
    else:
        return finley.fetch(rel_fn)


def image_pos(im, nph, prng):
    im[im < 0.0] = 0.0
    im = im.T/im.sum()
    idxs = prng.choice(im.size, size=nph, p=im.flatten())
    x, y = np.unravel_index(idxs, im.shape)
    dx = prng.uniform(low=0.5, high=1.5, size=x.size)
    dy = prng.uniform(low=0.5, high=1.5, size=y.size)
    return x+dx, y+dy


def find_nearest(a, b):
    return np.argmin(np.abs(a[:, np.newaxis] - b), axis=0)


def set_soxs_config(option, value):
    """
    Set SOXS configuration values.

    Parameters
    ----------
    option : string
        The option to change.
    value : number or string
        The value to set the option to.
    """
    from soxs.background.foreground import make_frgnd_spectrum
    bkgnd_options = ["abund_table", "apec_vers", "bkgnd_nH",
                     "bkgnd_absorb_model", "frgnd_spec_model"]
    soxs_cfg.set("soxs", option, value=value)
    if option in bkgnd_options:
        make_frgnd_spectrum()


def set_mission_config(mission):
    """
    Set configuration options most appropriate for a specific
    *mission*. Currently only takes "lem".
    """
    if mission == "lem":
        spec_model = "halosat"
    else:
        raise RuntimeError(f"Mission '{mission}' is not implemented!")
    set_soxs_config("frgnd_spec_model", spec_model)



import numpy as np
from soxs.constants import one_arcsec
from soxs.utils import parse_prng, parse_value, \
    get_rot_mat
import astropy.units as u
import astropy.wcs as pywcs
import astropy.io.fits as pyfits


def construct_wcs(ra0, dec0, dtheta=None, nx=None):
    if dtheta is None:
        dtheta = one_arcsec
    if nx is None:
        crpix = [0.0]*2
    else:
        crpix = [0.5*(nx+1)]*2
    w = pywcs.WCS(naxis=2)
    w.wcs.crval = [ra0, dec0]
    w.wcs.crpix = crpix
    w.wcs.cdelt = [-dtheta, dtheta]
    w.wcs.ctype = ["RA---TAN","DEC--TAN"]
    w.wcs.cunit = ["deg"]*2
    return w


def generate_radial_events(num_events, func, prng, ellipticity=1.0):
    rbins = np.linspace(0.0, 3000.0, 100000)
    rmid = 0.5*(rbins[1:]+rbins[:-1])
    pdf = func(rmid)*rmid
    pdf /= pdf.sum()
    radius = prng.choice(rmid, size=num_events, p=pdf)
    theta = 2.*np.pi*prng.uniform(size=num_events)
    x = radius*np.cos(theta)
    y = radius*np.sin(theta)*ellipticity
    return x, y


def rotate_xy(theta, x, y, inverse=False):
    rot_mat = get_rot_mat(theta)
    if inverse:
        rot_mat = np.linalg.inv(rot_mat)
    coords = np.dot(rot_mat, np.array([x, y]))
    return coords


def gen_img_coords(width, nx, theta):
    x, y = np.mgrid[0:nx,0:nx]-0.5*(nx-1)
    x *= width*60.0 / nx
    y *= width*60.0 / nx
    coords = rotate_xy(theta, x.flatten(), y.flatten(), inverse=True)
    return coords


class SpatialModel:
    def __init__(self, ra0, dec0):
        self.ra0 = parse_value(ra0, "deg")
        self.dec0 = parse_value(dec0, "deg")
        self.w = construct_wcs(self.ra0, self.dec0)

    def _generate_coords(self, num_events, prng):
        pass

    def _generate_image(self, width, nx):
        pass

    def generate_image(self, width, nx):
        """
        Generate an ImageHDU from the :class:`~soxs.spatial.SpatialModel`.

        Parameters
        ----------
        width : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The width of the image in arcminutes.
        nx : integer
            The resolution of the image, e.g. the number of pixels
            on a side.
        """
        width = parse_value(width, "arcmin")
        img = self._generate_image(width, nx)
        img /= img.sum()
        dtheta = width/60.0/nx
        w = construct_wcs(0.0, 0.0, dtheta=dtheta, nx=nx)
        imhdu = pyfits.ImageHDU(data=img.T, header=w.to_header())
        imhdu.name = "IMAGE"
        return imhdu

    def generate_coords(self, num_events, prng=None):
        """
        Generate a sample of photon positions from this
        spatial model.

        Parameters
        ----------
        num_events : integer
            The number of events to generate.
        prng : :class:`~numpy.random.RandomState` object, integer, or None
            A pseudo-random number generator. Typically will only
            be specified if you have a reason to generate the same
            set of random numbers, such as for a test. Default is None,
            which sets the seed based on the system time.
        """
        prng = parse_prng(prng)
        x, y = self._generate_coords(num_events, prng)
        ra, dec = self.w.wcs_pix2world(x, y, 1)
        return u.Quantity(ra, "deg"), u.Quantity(dec, "deg")


class PointSourceModel(SpatialModel):
    """
    A model for positions of photons emanating from 
    a point source.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The RA of the source in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The Dec of the source in degrees.
    """
    def __init__(self, ra0, dec0):
        super(PointSourceModel, self).__init__(ra0, dec0)

    def _generate_coords(self, num_events, prng):
        return (np.zeros(num_events),)*2
 
    def _generate_image(self, width, nx):
        img = np.zeros((nx,nx))
        img[nx//2, nx//2] = 1.0
        return img


class RadialFunctionModel(SpatialModel):
    """
    A model for positions of photons using a generic 
    surface brightness profile as a function of 
    radius.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the source in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the source in degrees.
    func : function or function-like, something callable.
        A function that takes an array of radii 
        and generates a radial surface brightness profile. 
    num_events : integer
        The number of events to generate. 
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model 
        in degrees. Only makes sense if ellipticity is 
        added. Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter 
        will shrink or expand the profile in the direction 
        of the "y" coordinate, so you may need to rotate 
        to get the shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, func, theta=0.0, ellipticity=1.0):
        super(RadialFunctionModel, self).__init__(ra0, dec0)
        self.theta = parse_value(theta, "deg")
        self.func = func
        self.ellipticity = ellipticity

    def _generate_coords(self, num_events, prng):
        x, y = generate_radial_events(num_events, self.func, prng,
                                      ellipticity=self.ellipticity)
        coords = rotate_xy(self.theta, x, y)
        return coords[0,:], coords[1,:]

    def _generate_image(self, width, nx):
        coords = gen_img_coords(width, nx, self.theta)
        coords[1,] /= self.ellipticity
        r = np.sqrt((coords**2).sum(axis=0))
        return self.func(r.reshape(nx,nx))


class RadialArrayModel(RadialFunctionModel):
    """
    Create positions for photons using a table of radii and 
    surface brightness contained in two arrays. 

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the source in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the source in degrees.
    r : NumPy array
        The array of radii for the profile in arcseconds. 
    S_r: NumPy array
        The array of the surface brightness of the profile. 
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model 
        in degrees. Only makes sense if ellipticity is 
        added. Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter will 
        shrink or expand the profile in the direction of the 
        "y" coordinate, so you may need to rotate to get the 
        shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, r, S_r, theta=0.0, ellipticity=1.0):
        func = lambda rr: np.interp(rr, r, S_r, left=0.0, right=0.0)
        super(RadialArrayModel, self).__init__(ra0, dec0, func, theta=theta, 
                                               ellipticity=ellipticity)


class RadialFileModel(RadialArrayModel):
    """
    A model for positions of photons using a table of radii and 
    surface brightness contained in a file. 

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the source in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the source in degrees.
    radfile : string
        The file containing the table of radii and surface 
        brightness. It must be an ASCII table with only two
        columns. 
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model in 
        degrees. Only makes sense if ellipticity is added. 
        Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter will 
        shrink or expand the profile in the direction of the 
        "y" coordinate, so you may need to rotate to get the 
        shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, radfile, theta=0.0, ellipticity=1.0):
        r, S_r = np.loadtxt(radfile, unpack=True)
        super(RadialFileModel, self).__init__(ra0, dec0, r, S_r, theta=theta, 
                                              ellipticity=ellipticity)


class BetaModel(RadialFunctionModel):
    """
    A model for positions of photons with a beta-model shape.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the beta model in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the beta model in degrees.
    r_c: float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The core radius of the profile in arcseconds.
    beta : float
        The "beta" parameter of the profile.
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model in 
        degrees. Only makes sense if ellipticity is added. 
        Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter will 
        shrink or expand the profile in the direction of the 
        "y" coordinate, so you may need to rotate to get the 
        shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, r_c, beta, theta=0.0, 
                 ellipticity=1.0):
        r_c = parse_value(r_c, "arcsec")
        func = lambda r: (1.0+(r/r_c)**2)**(-3*beta+0.5)
        super(BetaModel, self).__init__(ra0, dec0, func, theta=theta, 
                                        ellipticity=ellipticity)


class DoubleBetaModel(RadialFunctionModel):
    """
    A model for positions of photons with a double beta-model shape.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the beta model in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the beta model in degrees.
    r_c1: float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The inner core radius of the profile in arcseconds.
    beta1 : float
        The inner "beta" parameter of the profile.
    r_c2: float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The outer core radius of the profile in arcseconds.
    beta2 : float
        The outer "beta" parameter of the profile.
    sb_ratio : float
        The ratio of the outer to the inner SB peak value
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model in 
        degrees. Only makes sense if ellipticity is added. 
        Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter will 
        shrink or expand the profile in the direction of the 
        "y" coordinate, so you may need to rotate to get the 
        shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, r_c1, beta1, r_c2, beta2, sb_ratio,
                 theta=0.0, ellipticity=1.0):
        r_c1 = parse_value(r_c1, "arcsec")
        r_c2 = parse_value(r_c2, "arcsec")
        func = lambda r: (1.0+(r/r_c1)**2)**(-3*beta1+0.5) + \
                         sb_ratio*(1.0+(r/r_c2)**2)**(-3*beta2+0.5)
        super(DoubleBetaModel, self).__init__(ra0, dec0, func, theta=theta,
                                              ellipticity=ellipticity)


class AnnulusModel(RadialFunctionModel):
    """
    A model for positions of photons within an annulus shape 
    with uniform surface brightness.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the annulus in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the annulus in degrees.
    r_in : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The inner radius of the annulus in arcseconds.
    r_out: float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The outer radius of the annulus in arcseconds.
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the beta model in 
        degrees. Only makes sense if ellipticity is added. 
        Default: 0.0
    ellipticity : float, optional
        The ellipticity of the radial profile, expressed 
        as the ratio between the length scales of the x 
        and y coordinates. The value of this parameter will 
        shrink or expand the profile in the direction of the 
        "y" coordinate, so you may need to rotate to get the 
        shape you want. Default: 1.0
    """
    def __init__(self, ra0, dec0, r_in, r_out, theta=0.0, 
                 ellipticity=1.0):
        r_in = parse_value(r_in, "arcsec")
        r_out = parse_value(r_out, "arcsec")
        def func(r):
            f = np.zeros(r.shape)
            idxs = np.logical_and(r >= r_in, r < r_out)
            f[idxs] = 1.0
            return f
        super(AnnulusModel, self).__init__(ra0, dec0, func, 
                                           theta=theta,
                                           ellipticity=ellipticity)


class RectangleModel(SpatialModel):
    """
    A model for positions of photons within a rectangle 
    or line shape.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the rectangle in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the rectangle in degrees.
    width : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The width of the rectangle in arcseconds.
    height : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The height of the rectangle in arcseconds.
    theta : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The angle through which to rotate the rectangle 
        in degrees. Default: 0.0
    """
    def __init__(self, ra0, dec0, width, height, theta=0.0):
        super(RectangleModel, self).__init__(ra0, dec0)
        self.width = parse_value(width, "arcsec")
        self.height = parse_value(height, "arcsec")
        self.theta = parse_value(theta, "deg")

    def _generate_image(self, width, nx):
        img = np.zeros((nx, nx))
        c = gen_img_coords(width, nx, self.theta)
        rect = (-0.5*self.width < c[0]) & (c[0] < 0.5*self.width)
        rect &= (-0.5*self.height < c[1]) & (c[1] < 0.5*self.height)
        img[rect.reshape(nx,nx)] = 1.0
        return img

    def _generate_coords(self, num_events, prng):
        x = prng.uniform(low=-0.5*self.width, high=0.5*self.width, size=num_events)
        y = prng.uniform(low=-0.5*self.height, high=0.5*self.height, size=num_events)
        coords = rotate_xy(self.theta, x, y)
        return coords[0,:], coords[1,:]


class FillFOVModel(RectangleModel):
    """
    A model for positions of photons which span a field of view.

    Parameters
    ----------
    ra0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center RA of the field of view in degrees.
    dec0 : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The center Dec of the field of view in degrees.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The width of the field of view in arcminutes.
    """
    def __init__(self, ra0, dec0, fov):
        fov = parse_value(fov, "arcmin")
        width = fov*60.0
        height = fov*60.0
        super(FillFOVModel, self).__init__(ra0, dec0, width, height)



import numpy as np
from soxs.constants import keV_per_erg, erg_per_keV
from soxs.simput import SimputCatalog, SimputPhotonList
from soxs.spectra import get_wabs_absorb, get_tbabs_absorb
from soxs.utils import mylog, parse_prng, parse_value, soxs_cfg
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.special import erf
from astropy.table import Table
from astropy.io import ascii

# Function for computing spectral index of AGN sources
# a fit to the data from Figure 13a of Hickox & Markevitch 2006
# http://adsabs.harvard.edu/abs/2006ApJ...645...95H

# Parameters

aa = -14.0
bb = 0.5
cc = 0.5
dd = 1.8

# Here x = log10(flux)

def get_agn_index(x):
    y = (x-aa)/bb
    return cc*erf(y)+dd

# Index for galaxies

gal_index = 2.0

fb_emin = 0.5  # keV, low energy bound for the logN-logS flux band
fb_emax = 2.0  # keV, high energy bound for the logN-logS flux band

spec_emin = 0.1  # keV, minimum energy of mock spectrum
spec_emax = 10.0  # keV, max energy of mock spectrum


def get_flux_scale(ind, fb_emin, fb_emax, spec_emin, spec_emax):
    f_g = np.log(spec_emax/spec_emin)*np.ones(ind.size)
    f_E = np.log(fb_emax/fb_emin)*np.ones(ind.size)
    n1 = ind != 1.0
    n2 = ind != 2.0
    f_g[n1] = (spec_emax**(1.0-ind[n1])-spec_emin**(1.0-ind[n1]))/(1.0-ind[n1])
    f_E[n2] = (fb_emax**(2.0-ind[n2])-fb_emin**(2.0-ind[n2]))/(2.0-ind[n2])
    fscale = f_g/f_E
    return fscale


def generate_fluxes(fov, prng):
    from soxs.data import cdf_fluxes, cdf_gal, cdf_agn
    prng = parse_prng(prng)

    fov = parse_value(fov, "arcmin")

    logf = np.log10(cdf_fluxes)

    n_gal = np.rint(cdf_gal[-1])
    n_agn = np.rint(cdf_agn[-1])
    F_gal = cdf_gal / cdf_gal[-1]
    F_agn = cdf_agn / cdf_agn[-1]
    f_gal = InterpolatedUnivariateSpline(F_gal, logf)
    f_agn = InterpolatedUnivariateSpline(F_agn, logf)

    fov_area = fov**2

    n_gal = int(n_gal*fov_area/3600.0)
    n_agn = int(n_agn*fov_area/3600.0)
    mylog.debug(f"{n_agn} AGN, {n_gal} galaxies in the FOV.")

    randvec1 = prng.uniform(size=n_agn)
    agn_fluxes = 10**f_agn(randvec1)

    randvec2 = prng.uniform(size=n_gal)
    gal_fluxes = 10**f_gal(randvec2)

    return agn_fluxes, gal_fluxes


def generate_positions(num, fov, sky_center, prng):
    dec_scal = np.fabs(np.cos(sky_center[1] * np.pi / 180))
    ra_min = sky_center[0] - fov / (2.0 * 60.0 * dec_scal)
    dec_min = sky_center[1] - fov / (2.0 * 60.0)

    ra0 = prng.uniform(size=num) * fov / (60.0 * dec_scal) + ra_min
    dec0 = prng.uniform(size=num) * fov / 60.0 + dec_min

    return ra0, dec0


def generate_sources(fov, sky_center, prng=None):
    r"""
    Make a catalog of point sources.

    Parameters
    ----------
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    prng = parse_prng(prng)

    fov = parse_value(fov, "arcmin")

    agn_fluxes, gal_fluxes = generate_fluxes(fov, prng)

    fluxes = np.concatenate([agn_fluxes, gal_fluxes])

    ind = np.concatenate([get_agn_index(np.log10(agn_fluxes)),
                          gal_index * np.ones(gal_fluxes.size)])

    ra0, dec0 = generate_positions(fluxes.size, fov, sky_center, prng)

    return ra0, dec0, fluxes, ind


def make_ptsrc_background(exp_time, fov, sky_center, absorb_model=None, 
                          nH=None, area=40000.0, input_sources=None, 
                          output_sources=None, prng=None):
    r"""
    Make a point-source background.

    Parameters
    ----------
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time of the observation in seconds.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    absorb_model : string, optional
        The absorption model to use, "wabs" or "tbabs".
        Defaults to the value in the SOXS configuration file.
    nH : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The hydrogen column in units of 10**22 atoms/cm**2. 
        Defaults to the value in the SOXS configuration file.
    area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The effective area in cm**2. It must be large enough 
        so that a sufficiently large sample is drawn for the 
        ARF. Default: 40000.
    input_sources : string, optional
        If set to a filename, input the source positions, fluxes,
        and spectral indices from an ASCII table instead of generating
        them. Default: None
    output_sources : string, optional
        If set to a filename, output the properties of the sources
        within the field of view to a file. Default: None
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    prng = parse_prng(prng)

    exp_time = parse_value(exp_time, "s")
    fov = parse_value(fov, "arcmin")
    if nH is not None:
        nH = parse_value(nH, "1.0e22*cm**-2")
    area = parse_value(area, "cm**2")
    if input_sources is None:
        ra0, dec0, fluxes, ind = generate_sources(fov, sky_center, prng=prng)
        num_sources = fluxes.size
    else:
        mylog.info(f"Reading in point-source properties from {input_sources}.")
        t = ascii.read(input_sources)
        ra0 = t["RA"].data
        dec0 = t["Dec"].data
        fluxes = t["flux_0.5_2.0_keV"].data
        ind = t["index"].data
        num_sources = fluxes.size

    mylog.debug(f"Generating spectra from {num_sources} sources.")

    # If requested, output the source properties to a file
    if output_sources is not None:
        t = Table([ra0, dec0, fluxes, ind],
                  names=('RA', 'Dec', 'flux_0.5_2.0_keV', 'index'))
        t["RA"].unit = "deg"
        t["Dec"].unit = "deg"
        t["flux_0.5_2.0_keV"].unit = "erg/(cm**2*s)"
        t["index"].unit = ""
        t.write(output_sources, format='ascii.ecsv', overwrite=True)

    # Pre-calculate for optimization
    eratio = spec_emax/spec_emin
    oma = 1.0-ind
    invoma = 1.0/oma
    invoma[oma == 0.0] = 1.0
    fac1 = spec_emin**oma
    fac2 = spec_emax**oma-fac1

    fluxscale = get_flux_scale(ind, fb_emin, fb_emax, spec_emin, spec_emax)

    # Using the energy flux, determine the photon flux by simple scaling
    ref_ph_flux = fluxes*fluxscale*keV_per_erg
    # Now determine the number of photons we will generate
    n_photons = prng.poisson(ref_ph_flux*exp_time*area)

    all_energies = []
    all_ra = []
    all_dec = []

    for i, nph in enumerate(n_photons):
        if nph > 0:
            # Generate the energies in the source frame
            u = prng.uniform(size=nph)
            if ind[i] == 1.0:
                energies = spec_emin*(eratio**u)
            else:
                energies = fac1[i] + u*fac2[i]
                energies **= invoma[i]
            # Assign positions for this source
            ra = ra0[i]*np.ones(nph)
            dec = dec0[i]*np.ones(nph)

            all_energies.append(energies)
            all_ra.append(ra)
            all_dec.append(dec)

    mylog.debug("Finished generating spectra.")

    all_energies = np.concatenate(all_energies)
    all_ra = np.concatenate(all_ra)
    all_dec = np.concatenate(all_dec)

    # Remove some of the photons due to Galactic foreground absorption.
    # We will throw a lot of stuff away, but this is more general and still
    # faster.
    if nH is None:
        nH = float(soxs_cfg.get("soxs","bkgnd_nH"))
    if nH > 0.0:
        if absorb_model is None:
            absorb_model = soxs_cfg.get("soxs", "bkgnd_absorb_model")
        if absorb_model == "wabs":
            absorb = get_wabs_absorb(all_energies, nH)
        elif absorb_model == "tbabs":
            absorb = get_tbabs_absorb(all_energies, nH)
        randvec = prng.uniform(size=all_energies.size)
        all_energies = all_energies[randvec < absorb]
        all_ra = all_ra[randvec < absorb]
        all_dec = all_dec[randvec < absorb]
        all_nph = all_energies.size
        mylog.debug(f"{all_nph} photons remain after foreground galactic absorption.")

    all_flux = np.sum(all_energies)*erg_per_keV/(exp_time*area)

    output_events = {"ra": all_ra, "dec": all_dec, 
                     "energy": all_energies, "flux": all_flux}

    return output_events


def make_point_sources_file(filename, name, exp_time, fov, 
                            sky_center, absorb_model=None, nH=None,
                            area=40000.0, prng=None, append=False,
                            overwrite=False, src_filename=None,
                            input_sources=None, output_sources=None):
    """
    Make a SIMPUT catalog made up of contributions from
    point sources. 

    Parameters
    ----------
    filename : string
        The filename for the SIMPUT catalog.
    name : string
        The name of the SIMPUT photon list.
    exp_time : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The exposure time of the observation in seconds.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    absorb_model : string, optional
        The absorption model to use, "wabs" or "tbabs".
        Defaults to the value in the SOXS configuration file.
    nH : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The hydrogen column in units of 10**22 atoms/cm**2. 
        Defaults to the value in the SOXS configuration file.
    area : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
        The effective area in cm**2. It must be large enough 
        so that a sufficiently large sample is drawn for the 
        ARF. Default: 40000.
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time.
    append : boolean, optional
        If True, the photon list source will be appended to an existing
        SIMPUT catalog. Default: False
    overwrite : boolean, optional
        Set to True to overwrite previous files. Default: False
    src_filename : string, optional
        If set, this will be the filename to write the source
        to. By default, the source will be written to the same
        file as the SIMPUT catalog.
    input_sources : string, optional
        If set to a filename, input the source positions, fluxes,
        and spectral indices from an ASCII table instead of generating
        them. Default: None
    output_sources : string, optional
        If set to a filename, output the properties of the sources
        within the field of view to a file. Default: None
    """
    events = make_ptsrc_background(exp_time, fov, sky_center, 
                                   absorb_model=absorb_model, nH=nH, 
                                   area=area, input_sources=input_sources, 
                                   output_sources=output_sources, prng=prng)
    phlist = SimputPhotonList(events["ra"], events["dec"], events["energy"],
                              events["flux"], name=name)
    if append:
        cat = SimputCatalog.from_file(filename)
        cat.append(phlist, src_filename=src_filename, overwrite=overwrite)
    else:
        cat = SimputCatalog.from_source(filename, phlist, 
                                        src_filename=src_filename, 
                                        overwrite=overwrite)
    return cat


def make_point_source_list(output_file, fov, sky_center, prng=None):
    r"""
    Make a list of point source properties and write it to an ASCII
    table file.

    Parameters
    ----------
    output_file : string
        The ASCII table file to write the source properties to.
    fov : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
        The field of view in arcminutes.
    sky_center : array-like
        The center RA, Dec of the field of view in degrees.
    prng : :class:`~numpy.random.RandomState` object, integer, or None
        A pseudo-random number generator. Typically will only 
        be specified if you have a reason to generate the same 
        set of random numbers, such as for a test. Default is None, 
        which sets the seed based on the system time. 
    """
    ra0, dec0, fluxes, ind = generate_sources(fov, sky_center, prng=prng)

    t = Table([ra0, dec0, fluxes, ind],
              names=('RA', 'Dec', 'flux_0.5_2.0_keV', 'index'))
    t["RA"].unit = "deg"
    t["Dec"].unit = "deg"
    t["flux_0.5_2.0_keV"].unit = "erg/(cm**2*s)"
    t["index"].unit = ""
    t.write(output_file, format='ascii.ecsv', overwrite=True)



from soxs.utils import parse_prng, \
    parse_value, mylog, create_region, get_data_file
from soxs.background.events import make_diffuse_background
import numpy as np
import astropy.io.fits as pyfits
from regions import PixCoord


class InstrumentalBackground:

    def __init__(self, channel, count_rate, default_focal_length):
        self.channel = channel
        self.count_rate = count_rate
        self.default_focal_length = default_focal_length

    @classmethod
    def from_filename(cls, filename, ext_area, focal_length):
        """
        Read an instrumental background spectrum from 
        a FITS PHA file. 

        Parameters
        ----------
        filename : string
            The path to the file containing the spectrum.
        focal_length : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The default focal length of the instrument
            in meters. 
        """
        fn = get_data_file(filename)
        with pyfits.open(fn) as f:
            hdu = f["SPECTRUM"]
            if "COUNTS" in hdu.data.names:
                count_rate = hdu.data["COUNTS"]/hdu.header["EXPOSURE"]
            elif "COUNT_RATE" in hdu.data.names:
                count_rate = hdu.data["COUNT_RATE"]
            elif "RATE" in hdu.data.names:
                count_rate = hdu.data["RATE"]
            else:
                raise RuntimeError("Cannot find a field for either "
                                   "counts or count rate!")
            count_rate /= ext_area
            channel = hdu.data["CHANNEL"]
        return cls(channel, count_rate, focal_length)

    def generate_channel_spectrum(self, t_exp, solid_angle, 
                                  focal_length=None, prng=None):
        """
        Generate photon energy channels from this instrumental
        background spectrum given an exposure time,
        effective area, and solid angle.

        Parameters
        ----------
        t_exp : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The exposure time in seconds.
        solid_angle : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The solid angle in arcmin**2.
        focal_length : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
            The focal length in meters. Default is to use
            the default focal length of the instrument
            configuration.
        prng : :class:`~numpy.random.RandomState` object, integer, or None
            A pseudo-random number generator. Typically will only 
            be specified if you have a reason to generate the same 
            set of random numbers, such as for a test. Default is None, 
            which sets the seed based on the system time. 
        """
        t_exp = parse_value(t_exp, "s")
        solid_angle = parse_value(solid_angle, "arcmin**2")
        prng = parse_prng(prng)
        if focal_length is None:
            focal_length = self.default_focal_length
        else:
            focal_length = parse_value(focal_length, "m")
        fac = t_exp * solid_angle # Backgrounds are normalized to 1 arcmin**2
        fac *= (focal_length / self.default_focal_length) ** 2
        return prng.poisson(lam=self.count_rate*fac)

    def generate_channels(self, t_exp, solid_angle, focal_length=None,
                          prng=None):
        """
        Generate photon energy channels from this instrumental
        background spectrum given an exposure time,
        effective area, and solid angle.

        Parameters
        ----------
        t_exp : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The exposure time in seconds.
        solid_angle : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`
            The solid angle in arcmin**2.
        focal_length : float, (value, unit) tuple, or :class:`~astropy.units.Quantity`, optional
            The focal length in meters. Default is to use
            the default focal length of the instrument
            configuration.
        prng : :class:`~numpy.random.RandomState` object, integer, or None
            A pseudo-random number generator. Typically will only 
            be specified if you have a reason to generate the same 
            set of random numbers, such as for a test. Default is None, 
            which sets the seed based on the system time. 
        """
        ncts = self.generate_channel_spectrum(t_exp, solid_angle,
                                              focal_length=focal_length, 
                                              prng=prng)
        return np.concatenate([self.channel[i]*np.ones(n)
                               for i, n in enumerate(ncts)])


def make_instrument_background(inst_spec, event_params, rmf, prng=None):
    from collections import defaultdict

    prng = parse_prng(prng)

    bkgnd_spec = inst_spec["bkgnd"]

    if isinstance(bkgnd_spec[0], str):
        nchips = len(event_params["chips"])
        bkgnd_spec = [bkgnd_spec]*nchips

    bkg_events = defaultdict(list)
    pixel_area = (event_params["plate_scale"]*60.0)**2
    for i, chip in enumerate(event_params["chips"]):
        rtype = chip[0]
        args = chip[1:]
        r, bounds = create_region(rtype, args, 0.0, 0.0)
        sa = (bounds[1]-bounds[0])*(bounds[3]-bounds[2])*pixel_area
        bspec = InstrumentalBackground.from_filename(
            bkgnd_spec[i][0], bkgnd_spec[i][1], inst_spec['focal_length'])
        chan = bspec.generate_channels(
            event_params["exposure_time"], sa, prng=prng)
        n_events = chan.size
        detx = prng.uniform(low=bounds[0], high=bounds[1], size=n_events)
        dety = prng.uniform(low=bounds[2], high=bounds[3], size=n_events)
        if rtype in ["Box", "Rectangle"]:
            thisc = slice(None, None, None)
            n_det = n_events
        else:
            thisc = r.contains(PixCoord(detx, dety))
            n_det = thisc.sum()
        ch = chan[thisc].astype('int')
        e = rmf.ch_to_eb(ch, prng=prng)
        bkg_events["energy"].append(e)
        bkg_events[rmf.chan_type].append(ch)
        bkg_events["detx"].append(detx[thisc])
        bkg_events["dety"].append(dety[thisc])
        bkg_events["chip_id"].append(i*np.ones(n_det))
    for key in bkg_events:
        bkg_events[key] = np.concatenate(bkg_events[key])

    if bkg_events["energy"].size == 0:
        raise RuntimeError("No instrumental background events were detected!!!")
    else:
        mylog.info(f"Making {bkg_events['energy'].size} events "
                   f"from the instrumental background.")

    return make_diffuse_background(bkg_events, event_params, rmf, prng=prng)



import os
import shutil
import tempfile
from soxs.spectra import Spectrum
from soxs.spatial import PointSourceModel
from soxs.simput import SimputCatalog, SimputPhotonList
from soxs.instrument_registry import \
    make_simple_instrument
from soxs.instrument import instrument_simulator
from soxs.events import write_spectrum
from numpy.random import RandomState
from soxs.tests.utils import spectrum_answer_testing, \
    file_answer_testing

make_simple_instrument("lynx_hdxi", "new_hdxi", 20.0, 1024)

prng = RandomState(69)


def test_power_law(answer_store, answer_dir):
    plaw_fit(1.1, answer_store, answer_dir)
    plaw_fit(0.8, answer_store, answer_dir)
    plaw_fit(1.0, answer_store, answer_dir)


def plaw_fit(alpha_sim, answer_store, answer_dir):

    tmpdir = tempfile.mkdtemp()
    curdir = os.getcwd()
    os.chdir(tmpdir)

    nH_sim = 0.02
    norm_sim = 1.0e-4
    redshift = 0.01

    exp_time = (50.0, "ks")
    area = 40000.0
    inst_name = "new_hdxi"

    spec = Spectrum.from_powerlaw(alpha_sim, redshift, norm_sim, 0.1, 10.0, 
                                  20000)
    spec.apply_foreground_absorption(nH_sim, model="tbabs")

    spectrum_answer_testing(spec, f"power_law_{alpha_sim}.h5", answer_store, 
                            answer_dir)

    pt_src_pos = PointSourceModel(30.0, 45.0)
    pt_src = SimputPhotonList.from_models("plaw_model", spec, pt_src_pos,
                                           exp_time, area, prng=prng)
    cat = SimputCatalog.from_source("plaw_model_simput.fits", pt_src, 
                                    overwrite=True)

    instrument_simulator("plaw_model_simput.fits", 
                         f"plaw_model_{alpha_sim}_evt.fits",
                         exp_time, inst_name, [30.0, 45.0], instr_bkgnd=False,
                         ptsrc_bkgnd=False, foreground=False, prng=prng)

    write_spectrum(f"plaw_model_{alpha_sim}_evt.fits",
                   f"plaw_model_{alpha_sim}_evt.pha",
                   overwrite=True)

    file_answer_testing("EVENTS", f"plaw_model_{alpha_sim}_evt.fits",
                        answer_store, answer_dir)
    file_answer_testing("SPECTRUM", f"plaw_model_{alpha_sim}_evt.pha",
                        answer_store, answer_dir)

    os.chdir(curdir)
    shutil.rmtree(tmpdir)
